from datetime import datetime
import getpass
import logging

import git
import numpy as np
import xarray as xr

from indica.plasma import PlasmaProfiler
from indica.readers.modelreader import ModelReader
from indica.workflows.bda.bayesblackbox import BayesBlackBox
from indica.workflows.bda.optimisers import OptimiserContext
from indica.workflows.bda.priors import PriorManager


class BDADriver:
    def __init__(
        self,
        quant_to_optimise: list,
        opt_data: dict,
        plasma_profiler: PlasmaProfiler,
        prior_manager: PriorManager,
        modelreader: ModelReader,
        optimiser_context: OptimiserContext,
    ):
        self.quant_to_optimise = quant_to_optimise
        self.opt_data = opt_data
        self.plasma_profiler = plasma_profiler
        self.prior_manager = prior_manager
        self.modelreader = modelreader
        self.optimiser_context = optimiser_context

        self.blackbox = BayesBlackBox(
            opt_data=self.opt_data,
            quant_to_optimise=quant_to_optimise,
            ln_prior=self.prior_manager.ln_prior,
            plasma_profiler=self.plasma_profiler,
            build_bckc=self.modelreader.__call__,
        )

        self.optimiser_context.init_optimiser(self.blackbox.ln_posterior)

    def _build_inputs_dict(self):
        """

        Returns
        -------

        dictionary of inputs in MDS+ structure

        """

        result = {}
        quant_list = [item.split(".") for item in self.quant_to_optimise]

        result["ELEMENT"] = self.plasma_profiler.plasma.elements
        result["TIME"] = self.plasma_profiler.plasma.t.values
        git_id = git.Repo(search_parent_directories=True).head.object.hexsha

        result["INPUT"] = {
            "GIT_ID": f"{git_id}",
            "USER": f"{getpass.getuser()}",
            "SETTINGS": self.config,
            "DATETIME": datetime.utcnow().__str__(),
        }

        result["DIAG_DATA"] = {
            diag_name.upper(): {
                quantity[1].upper(): self.opt_data[f"{quantity[0]}.{quantity[1]}"]
                for quantity in quant_list
                if quantity[0] == diag_name
            }
            for diag_name in self.modelreader.models.keys()
        }

        return result

    def _build_result_dict(
        self,
    ):
        result = {}
        quant_list = [item.split(".") for item in self.quant_to_optimise]

        result["MODEL_DATA"] = {
            diag_name.upper(): {
                quantity[1].upper(): self.blobs[f"{quantity[0]}.{quantity[1]}"]
                for quantity in quant_list
                if quantity[0] == diag_name
            }
            for diag_name in self.modelreader.models.keys()
        }
        result["MODEL_DATA"]["SAMPLE_IDX"] = np.arange(
            0, self.opt_samples["post_sample"].shape[1]
        )

        result["PHANTOMS"] = {
            "FLAG": self.plasma_profiler.phantom,
            "PSI_NORM": {
                "RHOP": self.plasma_profiler.plasma.rhop,
                "NE": self.plasma_profiler.phantom_profiles["PSI_NORM"][
                    "electron_density"
                ],
                "TE": self.plasma_profiler.phantom_profiles["PSI_NORM"][
                    "electron_temperature"
                ],
                "TI": self.plasma_profiler.phantom_profiles["PSI_NORM"][
                    "ion_temperature"
                ],
                "NI": self.plasma_profiler.phantom_profiles["PSI_NORM"]["ion_density"],
                "NNEUTR": self.plasma_profiler.phantom_profiles["PSI_NORM"][
                    "neutral_density"
                ],
                "NFAST": self.plasma_profiler.phantom_profiles["PSI_NORM"][
                    "fast_ion_density"
                ],
                "ZEFF": self.plasma_profiler.phantom_profiles["PSI_NORM"]["zeff"].sum(
                    dim="element"
                ),
                "MEANZ": self.plasma_profiler.phantom_profiles["PSI_NORM"]["meanz"],
                "PTH": self.plasma_profiler.phantom_profiles["PSI_NORM"][
                    "thermal_pressure"
                ],
                "PFAST": self.plasma_profiler.phantom_profiles["PSI_NORM"][
                    "fast_ion_pressure"
                ],
                "P": self.plasma_profiler.phantom_profiles["PSI_NORM"]["pressure"],
                "VTOR": self.plasma_profiler.phantom_profiles["PSI_NORM"][
                    "toroidal_rotation"
                ],
            },
            "R_MIDPLANE": {
                "RPOS": self.plasma_profiler.plasma.R_midplane,
                "ZPOS": self.plasma_profiler.plasma.z_midplane,
                "NE": self.plasma_profiler.phantom_profiles["R_MIDPLANE"][
                    "electron_density"
                ],
                "TE": self.plasma_profiler.phantom_profiles["R_MIDPLANE"][
                    "electron_temperature"
                ],
                "TI": self.plasma_profiler.phantom_profiles["R_MIDPLANE"][
                    "ion_temperature"
                ],
                "NI": self.plasma_profiler.phantom_profiles["R_MIDPLANE"][
                    "ion_density"
                ],
                "NNEUTR": self.plasma_profiler.phantom_profiles["R_MIDPLANE"][
                    "neutral_density"
                ],
                "NFAST": self.plasma_profiler.phantom_profiles["R_MIDPLANE"][
                    "fast_ion_density"
                ],
                "ZEFF": self.plasma_profiler.phantom_profiles["R_MIDPLANE"]["zeff"].sum(
                    dim="element"
                ),
                "MEANZ": self.plasma_profiler.phantom_profiles["R_MIDPLANE"]["meanz"],
                "PTH": self.plasma_profiler.phantom_profiles["R_MIDPLANE"][
                    "thermal_pressure"
                ],
                "PFAST": self.plasma_profiler.phantom_profiles["R_MIDPLANE"][
                    "fast_ion_pressure"
                ],
                "P": self.plasma_profiler.phantom_profiles["R_MIDPLANE"]["pressure"],
                "VTOR": self.plasma_profiler.phantom_profiles["R_MIDPLANE"][
                    "toroidal_rotation"
                ],
            },
        }

        result["PROFILES"] = {
            "PSI_NORM": {
                "RHOP": self.plasma_profiler.plasma.rhop,
                "RHOT": self.plasma_profiler.plasma.equilibrium.rhot.interp(
                    t=self.plasma_profiler.plasma.t
                ),
                "VOLUME": self.plasma_profiler.plasma.volume,
                "NE": self.blobs["electron_density"].median(dim="sample_idx"),
                "NI": self.blobs["ion_density"].median(dim="sample_idx"),
                "TE": self.blobs["electron_temperature"].median(dim="sample_idx"),
                "TI": self.blobs["ion_temperature"].median(dim="sample_idx"),
                "NFAST": self.blobs["fast_ion_density"].median(dim="sample_idx"),
                "NNEUTR": self.blobs["neutral_density"].median(dim="sample_idx"),
                "P": self.blobs["pressure"].median(dim="sample_idx"),
                "PTH": self.blobs["thermal_pressure"].median(dim="sample_idx"),
                "PFAST": self.blobs["fast_ion_pressure"].median(dim="sample_idx"),
                "ZEFF": self.blobs["zeff"].sum("element").median(dim="sample_idx"),
                "MEANZ": self.blobs["meanz"].median(dim="sample_idx"),
                "NE_ERR": self.blobs["electron_density"].std(dim="sample_idx"),
                "NI_ERR": self.blobs["ion_density"].std(dim="sample_idx"),
                "TE_ERR": self.blobs["electron_temperature"].std(dim="sample_idx"),
                "TI_ERR": self.blobs["ion_temperature"].std(dim="sample_idx"),
                "NFAST_ERR": self.blobs["fast_ion_density"].std(dim="sample_idx"),
                "NNEUTR_ERR": self.blobs["neutral_density"].std(dim="sample_idx"),
                "P_ERR": self.blobs["pressure"].std(dim="sample_idx"),
                "PTH_ERR": self.blobs["thermal_pressure"].std(dim="sample_idx"),
                "PFAST_ERR": self.blobs["fast_ion_pressure"].std(dim="sample_idx"),
                "ZEFF_ERR": self.blobs["zeff"].sum("element").std(dim="sample_idx"),
                "MEANZ_ERR": self.blobs["meanz"].std(dim="sample_idx"),
                "VTOR": self.blobs["toroidal_rotation"].median(dim="sample_idx"),
                "VTOR_ERR": self.blobs["toroidal_rotation"].std(dim="sample_idx"),
            },
            "R_MIDPLANE": {
                "RPOS": self.plasma_profiler.plasma.R_midplane,
                "ZPOS": self.plasma_profiler.plasma.z_midplane,
                "NE": self.midplane_blobs["electron_density"].median(dim="sample_idx"),
                "NI": self.midplane_blobs["ion_density"].median(dim="sample_idx"),
                "TE": self.midplane_blobs["electron_temperature"].median(
                    dim="sample_idx"
                ),
                "TI": self.midplane_blobs["ion_temperature"].median(dim="sample_idx"),
                "NFAST": self.midplane_blobs["fast_ion_density"].median(
                    dim="sample_idx"
                ),
                "NNEUTR": self.midplane_blobs["neutral_density"].median(
                    dim="sample_idx"
                ),
                "P": self.midplane_blobs["pressure"].median(dim="sample_idx"),
                "PTH": self.midplane_blobs["thermal_pressure"].median(dim="sample_idx"),
                "PFAST": self.midplane_blobs["fast_ion_pressure"].median(
                    dim="sample_idx"
                ),
                "ZEFF": self.midplane_blobs["zeff"]
                .sum("element")
                .median(dim="sample_idx"),
                "MEANZ": self.midplane_blobs["meanz"].median(dim="sample_idx"),
                "NE_ERR": self.midplane_blobs["electron_density"].std(dim="sample_idx"),
                "NI_ERR": self.midplane_blobs["ion_density"].std(dim="sample_idx"),
                "TE_ERR": self.midplane_blobs["electron_temperature"].std(
                    dim="sample_idx"
                ),
                "TI_ERR": self.midplane_blobs["ion_temperature"].std(dim="sample_idx"),
                "NFAST_ERR": self.midplane_blobs["fast_ion_density"].std(
                    dim="sample_idx"
                ),
                "NNEUTR_ERR": self.midplane_blobs["neutral_density"].std(
                    dim="sample_idx"
                ),
                "P_ERR": self.midplane_blobs["pressure"].std(dim="sample_idx"),
                "PTH_ERR": self.midplane_blobs["thermal_pressure"].std(
                    dim="sample_idx"
                ),
                "PFAST_ERR": self.midplane_blobs["fast_ion_pressure"].std(
                    dim="sample_idx"
                ),
                "ZEFF_ERR": self.midplane_blobs["zeff"]
                .sum("element")
                .std(dim="sample_idx"),
                "MEANZ_ERR": self.midplane_blobs["meanz"].std(dim="sample_idx"),
                "VTOR": self.midplane_blobs["toroidal_rotation"].median(
                    dim="sample_idx"
                ),
                "VTOR_ERR": self.midplane_blobs["toroidal_rotation"].std(
                    dim="sample_idx"
                ),
            },
        }

        result["PROFILE_STAT"] = {
            "SAMPLE_IDX": np.arange(0, self.opt_samples["post_sample"].shape[1]),
            "PSI_NORM": {
                "RHOP": self.plasma_profiler.plasma.rhop,
                "NE": self.blobs["electron_density"],
                "NI": self.blobs["ion_density"],
                "TE": self.blobs["electron_temperature"],
                "TI": self.blobs["ion_temperature"],
                "NFAST": self.blobs["fast_ion_density"],
                "NNEUTR": self.blobs["neutral_density"],
                "P": self.blobs["pressure"],
                "PTH": self.blobs["thermal_pressure"],
                "PFAST": self.blobs["fast_ion_pressure"],
                "ZEFF": self.blobs["zeff"].sum("element"),
                "MEANZ": self.blobs["meanz"],
                "VTOR": self.blobs["toroidal_rotation"],
            },
            "R_MIDPLANE": {
                "RPOS": self.plasma_profiler.plasma.R_midplane,
                "NE": self.midplane_blobs["electron_density"],
                "NI": self.midplane_blobs["ion_density"],
                "TE": self.midplane_blobs["electron_temperature"],
                "TI": self.midplane_blobs["ion_temperature"],
                "NFAST": self.midplane_blobs["fast_ion_density"],
                "NNEUTR": self.midplane_blobs["neutral_density"],
                "P": self.midplane_blobs["pressure"],
                "PTH": self.midplane_blobs["thermal_pressure"],
                "PFAST": self.midplane_blobs["fast_ion_pressure"],
                "ZEFF": self.midplane_blobs["zeff"].sum("element"),
                "MEANZ": self.midplane_blobs["meanz"],
                "VTOR": self.midplane_blobs["toroidal_rotation"],
            },
        }

        result["OPTIMISATION"] = {
            "PRIOR_SAMPLE": self.opt_samples["prior_sample"],
            "POST_SAMPLE": self.opt_samples["post_sample"],
            "PARAM_NAMES": self.optimiser_context.optimiser_settings.param_names,
            "CONVERGENCE": self.convergence,
        }

        result["GLOBAL"] = {
            "VOLUME": self.plasma_profiler.plasma.volume.max(dim="rhop"),
            "TI0": self.blobs["ion_temperature"]
            .sel(rhop=0, method="nearest")
            .median(dim="sample_idx"),
            "TE0": self.blobs["electron_temperature"]
            .sel(rhop=0, method="nearest")
            .median(dim="sample_idx"),
            "NE0": self.blobs["electron_density"]
            .sel(rhop=0, method="nearest")
            .median(dim="sample_idx"),
            "NI0": self.blobs[
                "ion_density"
            ]  # TODO: where to concat the impurity_density onto this
            .sel(rhop=0, method="nearest")
            .median(dim="sample_idx"),
            "WP": self.blobs["wp"].median(dim="sample_idx"),
            "WTH": self.blobs["wth"].median(dim="sample_idx"),
            "ZEFF_AVG": self.midplane_blobs["zeff"]
            .sum(dim="element")
            .median(dim="sample_idx")
            .mean(dim="R"),
            "NNEUTR0": self.blobs["neutral_density"]
            .sel(rhop=0, method="nearest")
            .median(dim="sample_idx"),
            "NNEUTRB": self.blobs["neutral_density"]
            .sel(rhop=1, method="nearest")
            .median(dim="sample_idx"),
            "TI0_ERR": self.blobs["ion_temperature"]
            .sel(rhop=0, method="nearest")
            .std(dim="sample_idx"),
            "TE0_ERR": self.blobs["electron_temperature"]
            .sel(rhop=0, method="nearest")
            .std(dim="sample_idx"),
            "NE0_ERR": self.blobs["electron_density"]
            .sel(rhop=0, method="nearest")
            .std(dim="sample_idx"),
            "NI0_ERR": self.blobs["ion_density"]
            .sel(rhop=0, method="nearest")
            .std(dim="sample_idx"),
            "WTH_ERR": self.blobs["wth"].std(dim="sample_idx"),
            "WP_ERR": self.blobs["wp"].std(dim="sample_idx"),
            "ZEFF_AVG_ERR": self.midplane_blobs["zeff"]
            .sum(dim="element")
            .std(dim="sample_idx")
            .mean(dim="R"),
            "NNEUTR0_ERR": self.blobs["neutral_density"]
            .sel(rhop=0, method="nearest")
            .std(dim="sample_idx"),
            "NNEUTRB_ERR": self.blobs["neutral_density"]
            .sel(rhop=1, method="nearest")
            .std(dim="sample_idx"),
            "VTOR0": self.blobs["toroidal_rotation"]
            .sel(rhop=0, method="nearest")
            .median(dim="sample_idx"),
            "VTOR0_ERR": self.blobs["toroidal_rotation"]
            .sel(rhop=0, method="nearest")
            .std(dim="sample_idx"),
        }
        return result

    def __call__(
        self,
        config=None,
        time_points: np.ndarray = None,
    ):
        logger = logging.getLogger()
        self.config = config
        self.result = self._build_inputs_dict()
        results = []

        # TODO: problem with selecting time index and floating point precision errors
        if time_points is None:
            time_points = iter(self.plasma_profiler.plasma.t)

        for time in time_points:
            self.plasma_profiler.plasma.time_to_calculate = time
            logger.info(f"Time: {time.values:.2f}")
            self.optimiser_context.sample_start_points()
            self.optimiser_context.run()
            results.append(self.optimiser_context.post_process_results())
            self.optimiser_context.reset_optimiser()

        # unpack results and add time axis
        blobs = {}
        for key in results[0]["blobs"].keys():
            _blob = [result["blobs"][key] for result in results]
            blobs[key] = xr.concat(_blob, self.plasma_profiler.plasma.t)
        self.blobs = blobs
        self.midplane_blobs = self.plasma_profiler.map_plasma_profile_to_midplane(blobs)

        self.convergence = {}
        for key in results[0]["convergence"].keys():
            self.convergence[key] = [result["convergence"][key] for result in results]

        opt_samples = {}
        for key in results[0].keys():
            if key == "blobs":
                continue
            _opt_samples = [result[key] for result in results]
            opt_samples[key] = np.array(_opt_samples)
        self.opt_samples = opt_samples

        result = self._build_result_dict()
        self.result = dict(self.result, **result)
        return self.result



from typing import Dict
from typing import Optional
from typing import Tuple

import numpy as np
import xarray as xr
from xarray import apply_ufunc
from xarray import DataArray
from xarray import where

from indica.utilities import check_time_present
from .numpy_typing import FloatOrDataArray
from .numpy_typing import LabeledArray
from .numpy_typing import OnlyArray

_FLUX_TYPES = ["poloidal", "toroidal"]


class Equilibrium:
    """Class to hold and map equilibrium data.

    equilibrium_data
        Collection of equilibrium data
    R_shift
        Radial shift to test diagnostic mapping (positive for diagnostics)
        Either a float for all time slices or a DataArray with coord 't'
    z_shift
        Vertical z shift to test diagnostic mapping (positive for diagnostics)
        Either a float for all time slices or a DataArray with coord 't'

    TODO: should this class go in a sub-folder??
    """

    def __init__(
        self,
        equilibrium_data: Dict[str, DataArray],
        R_shift: FloatOrDataArray = 0.0,
        z_shift: FloatOrDataArray = 0.0,
    ):
        self.f: DataArray
        self.psi: DataArray
        self.psin: DataArray
        self.psi_boundary: DataArray
        self.psi_axis: DataArray
        self.ftor: DataArray
        self.rbnd: DataArray
        self.zbnd: DataArray
        self.rmag: DataArray
        self.zmag: DataArray

        self._data = equilibrium_data
        # Assign all equilibrium data as class attributes
        for k, v in equilibrium_data.items():
            setattr(self, k, v)

        # Substitute coordinates "psin" with "rhop"
        self.rhop = np.sqrt(self.psin)
        for k, v in equilibrium_data.items():
            if "psin" in v.dims:
                setattr(
                    self,
                    k,
                    v.assign_coords(rhop=("psin", self.rhop.data))
                    .swap_dims({"psin": "rhop"})
                    .drop_vars("psin"),
                )

        # Calculate volume and area if ajac and vjac are given
        if hasattr(self, "vjac") and hasattr(self, "ajac"):
            dpsin = self.psin[1] - self.psin[0]
            self.volume = (self.vjac * dpsin).cumsum("rhop")
            self.area = (self.ajac * dpsin).cumsum("rhop")

        # Calculate upper and lower boundaries
        self.zx_up = self.zbnd.min("index")
        self.zx_low = self.zbnd.max("index")

        # Calculate rho-toroidal and 2D rho matrix
        self.rhot = np.sqrt(
            (self.ftor - self.ftor.sel(rhop=0.0))
            / (self.ftor.sel(rhop=1.0) - self.ftor.sel(rhop=0.0))
        )
        self.rhop = np.sqrt(
            (self.psi - self.psi_axis) / (self.psi_boundary - self.psi_axis)
        )

        # TODO: shift of equilibrium is a bad idea, but useful...
        #   - psi (R, z) is restricted to new limits to avoid NaNs
        #   - volume is not changed so is inconsistent!
        if isinstance(R_shift, float):
            R_offset = xr.full_like(self.t, R_shift)
        else:
            R_offset = R_shift.interp(t=self.t, kwargs={"fill_value": 0})
        if isinstance(z_shift, float):
            z_offset = xr.full_like(self.t, z_shift)
        else:
            z_offset = z_shift.interp(t=self.t, kwargs={"fill_value": 0})

        self.R_offset = xr.where(np.isfinite(R_offset), R_offset, 0)
        self.z_offset = xr.where(np.isfinite(z_offset), z_offset, 0)
        if np.any(np.abs(R_offset) > 0) or np.any(np.abs(z_offset) > 0):
            R_new = self.psi.R + self.R_offset
            z_new = self.psi.z + self.z_offset
            R_range = slice(R_new.min("R"), R_new.max("R"))
            z_range = slice(z_new.min("z"), z_new.max("z"))
            self.psi = self.psi.interp(R=R_new, z=z_new).sel(R=R_range, z=z_range)
            self.rhop = self.rhop.interp(R=R_new, z=z_new).sel(R=R_range, z=z_range)
            self.rmag -= self.R_offset
            self.zmag -= self.z_offset
            self.rbnd -= self.R_offset
            self.zbnd -= self.z_offset
            self.zx_low -= self.z_offset
            if hasattr(self, "rmji"):
                self.rmji -= self.R_offset
            if hasattr(self, "rmjo"):
                self.rmjo -= self.R_offset

        if np.any(np.isnan(self.rhop)):
            self.rhop = xr.where(self.rhop > 0, self.rhop, 0.0)
        self.t = self.rhop.t
        self.Rmin = min(self.rhop.coords["R"])
        self.Rmax = max(self.rhop.coords["R"])
        self.zmin = min(self.rhop.coords["z"])
        self.zmax = max(self.rhop.coords["z"])
        self.corner_angles = [
            np.arctan2(self.zmin - self.zmag, self.Rmax - self.rmag) % (2 * np.pi),
            np.arctan2(self.zmax - self.zmag, self.Rmax - self.rmag) % (2 * np.pi),
            np.arctan2(self.zmax - self.zmag, self.Rmin - self.rmag) % (2 * np.pi),
            np.arctan2(self.zmin - self.zmag, self.Rmin - self.rmag) % (2 * np.pi),
        ]

    def Bfield(
        self,
        R: LabeledArray,
        z: LabeledArray,
        t: Optional[LabeledArray] = None,
        full_Rz: bool = False,
    ) -> Tuple[LabeledArray, LabeledArray, LabeledArray, LabeledArray]:
        """
        Magnetic field components at the desired time and location in space.

        R - Major radius position (m).
        z - The vertical position (m).
        t - Times (s).
        """

        if t is not None:
            check_time_present(t, self.t)
            psi = self.psi.interp(t=t, method="nearest", assume_sorted=True)
            f = self.f.interp(t=t, method="nearest", assume_sorted=True)
            _rhop, _, _ = self.flux_coords(R, z, t)
        else:
            t = self.rhop.coords["t"]
            psi = self.psi
            f = self.f
            _rhop, _, _ = self.flux_coords(R, z)

        dpsi_dR = psi.differentiate("R").interp(R=R, z=z)
        dpsi_dz = psi.differentiate("z").interp(
            R=R,
            z=z,
        )

        b_R = -(np.float64(1.0) / R) * dpsi_dz  # type: ignore
        b_R.name = "Radial magnetic field"
        b_R = b_R.T

        b_z = (np.float64(1.0) / R) * dpsi_dR  # type: ignore
        b_z.name = "Vertical Magnetic Field (T)"
        b_z = b_z.T
        _rhop = where(
            _rhop > np.float64(0.0), _rhop, np.float64(-1.0) * _rhop  # type: ignore
        )

        f = f.interp(rhop=_rhop)
        f.name = self.f.name
        b_T = f / R
        b_T.name = "Toroidal Magnetic Field (T)"

        if full_Rz:
            _b_T = b_T.interp(R=self.rmag, z=self.zmag) * self.rmag / self.rhop.R
            _b_T = _b_T.drop(["z", "rhop"]).expand_dims(dim={"z": self.rhop.z})
            b_T = _b_T.interp(R=R, z=z)

        return b_R, b_z, b_T, t

    def Btot(
        self,
        R: LabeledArray,
        z: LabeledArray,
        t: Optional[LabeledArray] = None,
        full_Rz: bool = False,
    ) -> Tuple[LabeledArray, LabeledArray]:
        """
        Total magnetic field at the desired time and location in space.

        R - Major radius position (m).
        z - The vertical position (m).
        t - Times (s).
        """
        b_R, b_z, b_T, t = self.Bfield(R, z, t, full_Rz=full_Rz)
        b_Tot = np.sqrt(
            b_R ** np.float64(2.0) + b_z ** np.float64(2.0) + b_T ** np.float64(2.0)
        )
        b_Tot.name = "Total Magnetic Field (T)"

        return b_Tot, t

    def Br(
        self, R: LabeledArray, z: LabeledArray, t: Optional[LabeledArray] = None
    ) -> Tuple[LabeledArray, LabeledArray]:
        """
        Radial magnetic field at the desired time and location in space.

        R - Major radius position (m).
        z - The vertical position (m).
        t - Times (s).
        """
        b_R, b_z, b_T, t = self.Bfield(R, z, t)

        return b_R, t

    def Bz(
        self, R: LabeledArray, z: LabeledArray, t: Optional[LabeledArray] = None
    ) -> Tuple[LabeledArray, LabeledArray]:
        """
        Vertical magnetic field at the desired time and location in space.

        R - Major radius position (m).
        z - The vertical position (m).
        t - Times (s).
        """
        b_R, b_z, b_T, t = self.Bfield(R, z, t)

        return b_z, t

    def Bt(
        self, R: LabeledArray, z: LabeledArray, t: Optional[LabeledArray] = None
    ) -> Tuple[LabeledArray, LabeledArray]:
        """
        Toroidal magnetic field at the desired time and location in space.

        R - Major radius position (m).
        z - The vertical position (m).
        t - Times (s).
        """
        b_R, b_z, b_T, t = self.Bfield(R, z, t)
        return b_T, t

    def Bp(
        self, R: LabeledArray, z: LabeledArray, t: Optional[LabeledArray] = None
    ) -> Tuple[LabeledArray, LabeledArray]:
        """
        Poloidal magnetic field at the desired time and location in space.

        R - Major radius position (m).
        z - The vertical position (m).
        t - Times (s).
        """
        b_R, b_z, b_T, t = self.Bfield(R, z, t)
        b_Pol = np.sqrt(b_R ** np.float64(2.0) + b_z ** np.float64(2.0))
        b_Pol.name = "Poloidal Magnetic Field (T)"
        return b_Pol, t

    def R_lfs(
        self,
        rho: LabeledArray,
        t: Optional[LabeledArray] = None,
        kind: str = "poloidal",
    ) -> Tuple[LabeledArray, LabeledArray]:
        """
        LFS major radius position of the given flux surface

        rho - Normalized flux coordinate values.
        t - Times (s).
        kind - Type of flux coordinate: "toroidal", "poloidal"
        """
        if t is None:
            rmjo = self.rmjo
            t = self.rmjo.coords["t"]
        else:
            check_time_present(t, self.t)
            rmjo = self.rmjo.interp(t=t, method="nearest")
        rhop, _ = self.convert_flux_coords(rho, t, from_kind=kind, to_kind="poloidal")
        R = rmjo.indica.interp2d(rhop=rhop, method="cubic")

        return R, t

    def R_hfs(
        self,
        rho: LabeledArray,
        t: Optional[LabeledArray] = None,
        kind: str = "poloidal",
    ) -> Tuple[LabeledArray, LabeledArray]:
        """
        HFS major radius position of the given flux surface

        rho - Normalized flux coordinate values.
        t - Times (s).
        kind - Type of flux coordinate: "toroidal", "poloidal"
        """
        if t is None:
            rmji = self.rmji
            t = self.rmji.coords["t"]
        else:
            check_time_present(t, self.t)
            rmji = self.rmji.interp(t=t, method="nearest")
        rhop, _ = self.convert_flux_coords(rho, t, from_kind=kind, to_kind="poloidal")
        R = rmji.indica.interp2d(rhop=rhop, method="cubic")

        return R, t

    def minor_radius(
        self,
        rho: LabeledArray,
        theta: LabeledArray,
        t: Optional[LabeledArray] = None,
        kind: str = "poloidal",
    ) -> Tuple[DataArray, LabeledArray]:
        """
        Minor radius of the given flux surface at a desired poloidal angle

        rho - Normalized flux coordinate values.
        theta - Poloidal angle.
        t - Times (s).
        kind - Type of flux coordinate: "toroidal", "poloidal"
        """
        ngrid = 100
        rhop, _ = self.convert_flux_coords(rho, t, from_kind=kind, to_kind="poloidal")
        theta = theta % (2 * np.pi)
        if t is not None:
            check_time_present(t, self.t)
            corner_angles = [
                angle.interp(t=t, method="nearest") for angle in self.corner_angles
            ]
            R0 = self.rmag.interp(t=t, method="nearest")
            z0 = self.zmag.interp(t=t, method="nearest")
            reference_rhos = self.rhop.interp(t=t, method="nearest")
            reference_rhos.name = self.rhop.name
        else:
            corner_angles = self.corner_angles
            R0 = self.rmag
            z0 = self.zmag
            reference_rhos = self.rhop
            t = self.rhop.coords["t"]
        minor_rad_max = apply_ufunc(
            lambda angle, corner1, corner2, corner3, corner4, R0, z0: (self.Rmax - R0)
            / np.cos(angle)
            if angle > corner1 or angle <= corner2
            else (self.zmax - z0) / np.sin(angle)
            if corner2 < angle <= corner3
            else (self.Rmin - R0) / np.cos(angle)
            if corner3 < angle <= corner4
            else (self.zmin - z0) / np.sin(angle),
            theta,
            corner_angles[0],
            corner_angles[1],
            corner_angles[2],
            corner_angles[3],
            R0,
            z0,
            vectorize=True,
        )
        minor_rads = DataArray(
            np.linspace(0.0, minor_rad_max, ngrid),
            dims=("r",) + minor_rad_max.dims,
            coords=minor_rad_max.coords,
        )
        R_grid = R0 + minor_rads * np.cos(theta)
        z_grid = z0 + minor_rads * np.sin(theta)
        fluxes_samples = reference_rhos.indica.interp2d(
            R=R_grid,
            z=z_grid,
            zero_coords={"R": R0, "z": z0},
            method="cubic",
            assume_sorted=True,
        ).rename("rhop")
        fluxes_samples.loc[{"r": 0}] = 0.0

        indices = fluxes_samples.indica.invert_root(rhop, "r", 0.0, method="cubic")
        return (
            minor_rads.indica.interp2d(r=indices, method="cubic", assume_sorted=True),
            t,
        )

    def flux_coords(
        self,
        R: OnlyArray,
        z: OnlyArray,
        t: Optional[LabeledArray] = None,
        kind: str = "poloidal",
    ) -> Tuple[DataArray, DataArray, LabeledArray]:
        """
        Normalised flux coordinate and angle at a given location in space.

        R - Major radius position (m).
        z - The vertical position (m).
        t - Times (s).
        kind - Type of flux coordinate: "toroidal", "poloidal"
        """

        if t is None:
            rhop = self.rhop
            R_ax = self.rmag
            z_ax = self.zmag
            t = self.rhop.coords["t"]
            z_x_point_low = self.zx_low
            z_x_point_up = self.zx_up
        else:
            check_time_present(t, self.t)
            rhop = self.rhop.interp(t=t, method="nearest")
            R_ax = self.rmag.interp(t=t, method="nearest")
            z_ax = self.zmag.interp(t=t, method="nearest")
            z_x_point_low = self.zx_low.interp(t=t, method="nearest")
            z_x_point_up = self.zx_up.interp(t=t, method="nearest")

        # TODO: rho and theta dimensions not in the same order...
        rhop = rhop.interp(R=R, z=z)
        theta = np.arctan2(
            z - z_ax,
            R - R_ax,
        )

        # Correct for any interpolation errors resulting in negative fluxes
        rhop = xr.where((rhop < 0.0) * (rhop > -1e-12), 0.0, rhop)

        if kind != "poloidal":
            rhop, t = self.convert_flux_coords(rhop, t, "poloidal", kind)

        # Set rho to be negative in the private flux region
        rhop = xr.where(
            (rhop < 1.0) * (z < z_x_point_low) * (z < z_x_point_up), -rhop, rhop
        )

        # Convert to desired normalised flux coordinate
        rho, _ = self.convert_flux_coords(rhop, t, from_kind="poloidal", to_kind=kind)

        return rho, theta, t

    def spatial_coords(
        self,
        rho: LabeledArray,
        theta: LabeledArray,
        t: Optional[LabeledArray] = None,
        kind: str = "poloidal",
    ) -> Tuple[LabeledArray, LabeledArray, LabeledArray]:
        """
        (R, z) coordinates of a flux surface given its (rho, theta).

        rho - Normalized flux coordinate values.
        theta - Poloidal angle.
        t - Times (s).
        kind - Type of flux coordinate: "toroidal", "poloidal"
        """
        rhop, _ = self.convert_flux_coords(rho, t, from_kind=kind, to_kind="poloidal")
        minor_rad, t = self.minor_radius(rhop, theta, t, kind)
        R0 = self.rmag.interp(t=t, method="nearest")
        z0 = self.zmag.interp(t=t, method="nearest")
        R = R0 + minor_rad * np.cos(theta)
        z = z0 + minor_rad * np.sin(theta)
        return R, z, t

    def convert_flux_coords(
        self,
        rho: LabeledArray,
        t: Optional[LabeledArray] = None,
        from_kind: Optional[str] = "poloidal",
        to_kind: Optional[str] = "toroidal",
    ) -> Tuple[LabeledArray, LabeledArray]:
        """
        Convert between normalized flux coordinates coordinate systems.

        rho - Normalized flux coordinate values.
        t - Times (s).
        from_kind - Input flux coordinate: "toroidal", "poloidal"
        to_kind - Output flux coordinate: "poloidal", "toroidal"
        """
        if from_kind == to_kind:
            return rho, t

        supported = ["poloidal", "toroidal"]
        if from_kind not in supported or to_kind not in supported:
            raise ValueError("kind must be either poloidal or toroidal")

        if t is not None:
            check_time_present(t, self.t)
            conversion = self.rhot.interp(t=t, method="nearest")
        else:
            conversion = self.rhot
            t = self.rhot.t

        if to_kind == "toroidal":
            _rho = conversion.indica.interp2d(
                rhop=np.abs(rho), method="cubic", assume_sorted=True
            )
        elif to_kind == "poloidal":
            _rho = conversion.indica.invert_interp(np.abs(rho), "rhop", method="cubic")

        return _rho, t

    def cross_sectional_area(
        self,
        rho: LabeledArray,
        t: Optional[LabeledArray] = None,
        kind: str = "poloidal",
    ) -> Tuple[DataArray, LabeledArray]:
        """
        Cross-sectional area of desired flux surface

        rho - Normalized flux coordinate values.
        t - Times (s).
        kind - Normalised flux coordinate type: "toroidal", "poloidal"
        """
        rhop, _ = self.convert_flux_coords(rho, t, from_kind=kind, to_kind="poloidal")
        if t is None:
            t = self.rhop.coords["t"]
        else:
            check_time_present(t, self.t)
        area = self.area.interp(rhop=rhop).interp(t=t)
        return area, t

    def enclosed_volume(
        self,
        rho: LabeledArray,
        t: Optional[LabeledArray] = None,
        kind: str = "poloidal",
    ) -> Tuple[DataArray, LabeledArray]:
        """
        Volume of desired flux surface

        rho - Normalized flux coordinate values.
        t - Times (s).
        kind - Normalised flux coordinate type: "toroidal", "poloidal"
        """
        rhop, _ = self.convert_flux_coords(rho, t, from_kind=kind, to_kind="poloidal")
        if t is None:
            t = self.rhop.coords["t"]
        else:
            check_time_present(t, self.t)
        volume = self.area.interp(rhop=rhop).interp(t=t)
        return volume, t

    def write_to_geqdsk(self):
        # TODO: Implement writing to geqdsk
        raise NotImplementedError("Method not yet implemented")



import numpy as np
import xarray as xr

from indica.profilers.profiler_base import ProfilerBase


class ProfilerBasis(ProfilerBase):
    """
    Class to build pca profiles from basis functions

    Parameters
    ----------
    radial_grid
        normalised radial grid [0, 1]  on which profile is to be built
    """

    def __init__(
        self,
        basis_functions: np.ndarray,
        bias: np.ndarray,
        ncomps: int = 2,
        radial_grid: np.ndarray = None,
        coord="rhop",
        parameters: dict = None,
    ):
        super().__init__(parameters)
        self.basis_functions = basis_functions
        self.bias = bias
        self.ncomps = ncomps
        self.radial_grid = radial_grid
        self.coord = coord

        # Weights have to be dynamically assigned as attributes, so they can be accessed
        # / changed by the profiler
        for icomp in range(self.ncomps):
            param_name = f"weight_{icomp + 1}"
            self.parameters[param_name] = 0
            setattr(self, param_name, 0)

    def construct_profile(
        self,
    ):
        weights = np.stack(
            [weight for weight_name, weight in self.parameters.items()], axis=-1
        ).T
        return np.dot(weights, self.basis_functions) + self.bias

    def __call__(
        self,
    ):
        """
        Builds the profile from basis functions using the parameters set
        """
        y = self.construct_profile()
        self.ydata = xr.DataArray(y, coords=[(self.coord, self.radial_grid.data)])
        return self.ydata



"""Callback functions for choosing which channels of data should be kept."""

import json
from numbers import Number
from typing import Callable
from typing import Collection
from typing import Dict
from typing import Hashable
from typing import Iterable
from typing import List
from typing import Optional

from matplotlib.backend_bases import PickEvent
from matplotlib.collections import PathCollection
import matplotlib.pyplot as plt
from xarray import DataArray

#: Callback function which can be used to choose which data should be read.
DataSelector = Callable[
    [DataArray, str, Collection[Number], Iterable[Number]], Iterable[Number]
]

MAX_TIME_SLICES = 10


class PickStack:
    def __init__(self, stack, on_pick):
        self.stack = stack
        self.ax = [artist.axes for artist in self.stack][0]
        self.on_pick = on_pick
        self.cid = self.ax.figure.canvas.mpl_connect(
            "button_press_event", self.fire_pick_event
        )

    def fire_pick_event(self, event):
        if not event.inaxes:
            return
        cont = [a for a in self.stack if a.contains(event)[0]]
        if not cont:
            return
        pick_event = PickEvent(
            "pick_Event",
            self.ax.figure.canvas,
            event,
            cont[0],
            guiEvent=event.guiEvent,
            **cont[0].contains(event)[1],
        )
        self.on_pick(pick_event)


def choose_on_plot(
    data: DataArray,
    channel_dim: str,
    bad_channels: Collection[Number],
    unselected_channels: Iterable[Number] = [],
) -> Iterable[Number]:
    """Produces a plot in its own window, on which the user can click data
    from channels which they wish to be discarded.

    Instructions are printed to the command line, so it is probably
    not suitable for use in a GUI.

    Parameters
    ----------
    data:
        The data from which channels should be selected to discard.
    channel_dim:
        The name of the dimension used for storing separate channels. This
        will be used for the x-axis in the plot.
    bad_channels:
        A (possibly empty) list of channel labels which are known to be
        incorrectly calibrated, faulty, or otherwise untrustworty. These will
        be plotted in red, but must still be specifically selected by the user
        to be discared.
    unselected_channels:
        A (possibly empty) list of channels which will be ignored by default.
        These could be, e.g., cached values from a previous run. The user is
        welcome to change this.

    Returns
    -------
    :
        A list of channel labels which the user has selected to be discarded.

    """
    print(
        "Click any channel you wish to ignore. Click a second time to "
        "include it\nagain. Ignored channels are in grey and suspicious "
        "channels are in red.\n"
    )
    print("Once finished, close the plot.\n")

    channels_to_drop = list(unselected_channels)

    # Get colour array
    original_colours = [
        "r" if label in bad_channels else "b" for label in data.coords[channel_dim]
    ]
    colours = [
        c if label not in channels_to_drop else "#ADADAD"
        for c, label in zip(original_colours, data.coords[channel_dim])
    ]

    plots: List[PathCollection] = []

    def on_pick(event):
        ind = event.ind[0]
        label = data.coords[channel_dim].data[ind]
        if label in channels_to_drop:
            channels_to_drop.remove(label)
            colours[ind] = original_colours[ind]
            print("Including data at {}={}".format(channel_dim, label))
        else:
            channels_to_drop.append(label)
            colours[ind] = "#ADADAD"
            print("Ignoring data at {}={}".format(channel_dim, label))
        for plot in plots:
            plot.set_color(colours)  # Does this work? Or should I use set_facecolor?
        # Not sure whether I'll need this:
        event.canvas.draw()

    if data.ndim > 2:
        raise ValueError("Received DataArray with more than 2 dimensions.")
    channel_pos = data.dims.index(channel_dim)
    if data.ndim > 1:
        other_dim: Optional[Hashable] = (
            data.dims[0] if channel_pos == 1 else data.dims[1]
        )
    else:
        other_dim = None

    if other_dim is None:
        plots.append(
            plt.scatter(data.coords[channel_dim], data, c=colours, picker=True)
        )
    else:
        # Work out which time slices to plot
        available_rows = len(data.coords[other_dim])
        stride = available_rows // MAX_TIME_SLICES
        for row in range(0, available_rows, stride):
            plots.append(
                plt.scatter(
                    data.coords[channel_dim],
                    data[{other_dim: row, channel_dim: slice(None)}],
                    c=colours,
                    picker=True,
                )
            )

    p = PickStack(plots, on_pick)
    plt.xlabel(channel_dim)
    datatype = data.attrs["datatype"]
    plt.ylabel(f"{datatype[1]} {datatype[0]}")
    plt.show()
    print("---------------------------------------------------------------------\n")
    del p
    return channels_to_drop


def use_cached_ignore_channels(
    data: DataArray,
    channel_dim: str,
    bad_channels: Collection[Number],
    unselected_channels: Iterable[Number] = [],
) -> Iterable[Number]:
    """
    Return channels from cache with no modification/input.
    """
    return list(unselected_channels)


def ignore_channels_from_dict(ignore_dict: Dict[str, List[Number]]) -> Callable:
    """
    Ignore channels from dictionary of {instrument: list of channel} pairs

    :param ignore_dict: Dictionary of channels to ignore per instrument
                        Use keys of {instrument}_{quantity}
    :return: Channel selection function to pass to reader
    """

    def ignore_channels(
        data: DataArray,
        channel_dim: str,
        bad_channels: Collection[Number],
        unselected_channels: Iterable[Number] = [],
    ) -> Iterable[Number]:
        return ignore_dict.get(str(data.name), [])

    return ignore_channels


def ignore_channels_from_file(filename: str) -> Callable:
    """
    Load channels to ignore from JSON. Assume file contains dict-like structure with
    instruments as keys and list of channels as values

    :param filename: Path to JSON file containing channels to ignore
    :return: Channel selection function to pass to reader
    """
    with open(file=filename, mode="r") as f:
        channels = json.load(f)
    return ignore_channels_from_dict(ignore_dict=channels)



"""Routines for averaging or interpolate along the time axis given start and end time
and a desired time resolution"""

import numpy as np
from xarray import DataArray


def convert_in_time(
    tstart: float,
    tend: float,
    frequency: float,
    data: DataArray,
    method: str = "linear",
) -> DataArray:
    """Interpolate or bin (as appropriate) the given data along the time
    axis, discarding data before or after the limits.

    Parameters
    ----------
    tstart
        The lower limit in time for determining which data to retain.
    tend
        The upper limit in time for determining which data to retain.
    frequency
        Frequency of sampling on the time axis.
    data
        Data to be interpolated/binned.
    method
        Interpolation method to use. Must be a value accepted by
        :py:class:`scipy.interpolate.interp1d`.

    Returns
    -------
    :
        Array like the input, but interpolated or binned along the time axis.

    """
    # TODO: Generate PROV information
    original_freq = 1 / (data.coords["t"][1] - data.coords["t"][0])
    if frequency / original_freq <= 0.2:
        return bin_in_time(tstart, tend, frequency, data)
    else:
        return interpolate_in_time(tstart, tend, frequency, data, method)


def convert_in_time_dt(
    tstart: float,
    tend: float,
    dt: float,
    data: DataArray,
    method: str = "linear",
) -> DataArray:
    """
    Interpolate or bin given data along the time axis, discarding data before
    or after the limits.

    Parameters
    ----------
    tstart
        The lower limit in time for determining which data to retain.
    tend
        The upper limit in time for determining which data to retain.
    dt
        Time resolution of new time axis.
    data
        Data to be interpolated/binned.

    Returns
    -------
    :
        Array like the input, but interpolated/binned along the time axis.

    """

    tcoords = data.coords["t"]
    data_dt = tcoords[1] - tcoords[0]
    if data_dt <= dt / 2 and tstart != tend:
        return bin_in_time_dt(tstart, tend, dt, data)
    else:
        return interpolate_in_time_dt(tstart, tend, dt, data, method=method)


def interpolate_to_time_labels(
    tlabels: np.ndarray, data: DataArray, method: str = "linear"
) -> DataArray:
    """
    Interpolate data to sit on the specified time labels.

    Parameters
    ----------
    tlabels
        The time at which the data should be interpolated.
    data
        Data to be interpolated.

    Returns
    -------
    :
        Array like the input, but interpolated onto the time labels.

    """
    if data.coords["t"].shape == tlabels.shape and np.all(data.coords["t"] == tlabels):
        return data

    interpolated = data.interp(t=tlabels, method=method)

    return interpolated


def bin_to_time_labels(tlabels: np.ndarray, data: DataArray) -> DataArray:
    """Bin data to sit on the specified time labels.

    Parameters
    ----------
    tlabels
        The time at which the data should be binned.
    data
        Data to be binned.

    Returns
    -------
    :
        Array like the input, but binned onto the time labels.

    """
    if data.coords["t"].shape == tlabels.shape and np.all(data.coords["t"] == tlabels):
        return data
    npoints = len(tlabels)
    half_interval = 0.5 * (tlabels[1] - tlabels[0])
    tbins = np.empty(npoints + 1)
    tbins[0] = tlabels[0] - half_interval
    tbins[1:] = tlabels + half_interval
    grouped = data.sel(t=slice(tbins[0], tbins[-1])).groupby_bins(
        "t", tbins, labels=tlabels
    )
    averaged = grouped.mean("t", keep_attrs=True, skipna=True)
    averaged = averaged.rename(t_bins="t")

    stdev = grouped.std("t", keep_attrs=True, skipna=True)
    stdev = np.sqrt(stdev**2)
    stdev = stdev.rename(t_bins="t")
    averaged = averaged.assign_coords(stdev=(data.dims, stdev.data))

    if "error" in data.coords:
        grouped = data.error.sel(t=slice(tbins[0], tbins[-1])).groupby_bins(
            "t", tbins, labels=tlabels
        )
        uncertainty = np.sqrt(
            grouped.reduce(
                lambda x, axis: np.sum(x**2, axis) / np.size(x, axis) ** 2, "t"
            )
        )
        error = np.sqrt(uncertainty**2)
        error = error.rename(t_bins="t")
        averaged = averaged.assign_coords(error=(data.dims, error.data))

    return averaged


def interpolate_in_time(
    tstart: float,
    tend: float,
    frequency: float,
    data: DataArray,
    method: str = "linear",
) -> DataArray:
    """Interpolate the given data along the time axis, discarding data
    before or after the limits.

    Parameters
    ----------
    tstart
        The lower limit in time for determining which data to retain.
    tend
        The upper limit in time for determining which data to retain.
    frequency
        Frequency of sampling on the time axis.
    data
        Data to be interpolated.
    method
        Interpolation method to use. Must be a value accepted by
        :py:class:`scipy.interpolate.interp1d`.

    Returns
    -------
    :
        Array like the input, but interpolated along the time axis.

    """
    check_bounds_interp(tstart, tend, data)

    tlabels = get_tlabels(tstart, tend, frequency)

    return interpolate_to_time_labels(tlabels, data, method=method)


def interpolate_in_time_dt(
    tstart: float,
    tend: float,
    dt: float,
    data: DataArray,
    method: str = "linear",
) -> DataArray:
    """Interpolate the given data along the time axis, discarding data
    before or after the limits.

    Parameters
    ----------
    tstart
        The lower limit in time for determining which data to retain.
    tend
        The upper limit in time for determining which data to retain.
    dt
        Time resolution of new time axis.
    data
        Data to be interpolated.
    method
        Interpolation method to use. Must be a value accepted by
        :py:class:`scipy.interpolate.interp1d`.

    Returns
    -------
    :
        Array like the input, but interpolated along the time axis.

    """

    check_bounds_interp(tstart, tend, data)
    tlabels = get_tlabels_dt(tstart, tend, dt)

    return interpolate_to_time_labels(tlabels, data, method=method)


def bin_in_time(
    tstart: float, tend: float, frequency: float, data: DataArray
) -> DataArray:
    """Bin given data along the time axis, discarding data before or after
    the limits.

    Parameters
    ----------
    tstart
        The lower limit in time for determining which data to retain.
    tend
        The upper limit in time for determining which data to retain.
    frequency
        Frequency of sampling on the time axis.
    data
        Data to be binned.

    Returns
    -------
    :
        Array like the input, but binned along the time axis.

    """
    check_bounds_bin(tstart, tend, 1.0 / frequency, data)
    tlabels = get_tlabels(tstart, tend, frequency)

    return bin_to_time_labels(tlabels, data)


def bin_in_time_dt(tstart: float, tend: float, dt: float, data: DataArray) -> DataArray:
    """Bin given data along the time axis, discarding data before or after
    the limits.

    Parameters
    ----------
    tstart
        The lower limit in time for determining which data to retain.
    tend
        The upper limit in time for determining which data to retain.
    dt
        Time resolution of new time axis.
    data
        Data to be binned.

    Returns
    -------
    :
        Array like the input, but binned along the time axis.
    TODO: add possibility of doing 50% overlap of time bins!
    """
    check_bounds_bin(tstart, tend, dt, data)
    tlabels = get_tlabels_dt(tstart, tend, dt)
    return bin_to_time_labels(tlabels, data)


def get_tlabels(tstart: float, tend: float, frequency: float):
    """
    Build time array given start, end and frequency

    Parameters
    ----------
    tstart
        The lower limit in time for determining which data to retain.
    tend
        The upper limit in time for determining which data to retain.
    frequency
        Frequency of sampling on the time axis.

    Returns
    -------
    tlabels
        Time array

    """
    npoints = round((tend - tstart) * frequency) + 1
    return np.linspace(tstart, tend, npoints)


def get_tlabels_dt(tstart: float, tend: float, dt: float):
    """
    Build time array given start, end and frequency

    Parameters
    ----------
    tstart
        The lower limit in time for determining which data to retain.
    tend
        The upper limit in time for determining which data to retain.
    dt
        Time resolution of new time axis.

    Returns
    -------
    tlabels
        Time array

    """
    _tlabels = np.arange(tstart, tend + dt, dt)
    tlabels = _tlabels[np.logical_or(_tlabels < tend, np.isclose(_tlabels, tend))]
    return np.array(tlabels, ndmin=1)


def check_bounds_bin(tstart: float, tend: float, dt: float, data: DataArray):
    """
    Check necessary bounds for binning data in time

    Parameters
    ----------
    tstart
        The lower limit in time for determining which data to retain.
    tend
        The upper limit in time for determining which data to retain.
    dt
        Time resolution of new time axis.
    data
        Data to be binned.
    """
    tcoords = data.coords["t"]
    half_interval = dt / 2
    if tcoords[0] > tstart + half_interval:
        raise ValueError(
            "No data falls within first bin {}.".format(
                (tstart - half_interval, tstart + half_interval)
            )
        )
    if tcoords[-1] < tend - half_interval:
        raise ValueError(
            "No data falls within last bin {}.".format(
                (tend - half_interval, tend + half_interval)
            )
        )


def check_bounds_interp(tstart: float, tend: float, data: DataArray):
    """
    Check necessary bounds for interpolating in time

    Parameters
    ----------
    tstart
        The lower limit in time for determining which data to retain.
    tend
        The upper limit in time for determining which data to retain.
    dt
        Time resolution of new time axis.
    data
        Data to be interpolated.
    """
    tcoords = data.coords["t"]
    start = np.argmax((tcoords > tstart).data) - 1
    if start < 0:
        raise ValueError("Start time {} not in range of provided data.".format(tstart))
    end = np.argmax((tcoords >= tend).data)
    if end < 1:
        raise ValueError("End time {} not in range of provided data.".format(tend))

    return


#
# def run_example(nt=50, plot=False):
#     values = np.sin(np.linspace(0, np.pi * 3, nt)) + np.random.random(nt) - 0.5
#     time = np.linspace(0, 0.1, nt)
#     data = DataArray(values, coords=[("t", time)])
#
#     dt = time[1] - time[0]
#     dt_binned = dt * 4
#     dt_interp = dt / 4
#
#     tstart = time[0] + 5 * dt
#     tend = time[-1] - 10 * dt
#     data_interp = convert_in_time_dt(tstart, tend, dt_interp, data)
#     data_binned = convert_in_time_dt(tstart, tend, dt_binned, data)
#
#     if plot:
#         import matplotlib.pylab as plt
#
#         plt.figure()
#         data_interp.plot(marker="x", label="Interpolated")
#         data.plot(marker="o", label="Original data")
#         data_binned.plot(marker="x", label="Binned")
#         plt.legend()
#         plt.show()



import os

import numpy as np
import xarray as xr

import indica

# Constants
RY = 13.605  # eV
PERCMTOEV = 1.239841e-4  # Convert 1/cm to eV

head = os.path.dirname(indica.__file__)
FILEHEAD = os.path.join(head, "data/Data_Argon/")

# TODO: worth renaming this file/methods (Marchuck -> Helike?)
#      and referencing where the data comes from


def diel_calc(atomic_data: np.typing.ArrayLike, Te: xr.DataArray, label: str = "he"):
    """
    Calculates intensity of dielectronic recombination
    Parameters
    ----------
    atomic_data
        array of atomic data read from Marchuk's database
    Te
        electron temperature (eV)
    label
        "he" for helium like collision or "li" for lithium like inner collision
    Returns
    -------
    Intensity along Te vector
    """
    a0 = 5.29e-9  # bohr radius / cm
    Te = Te / RY
    Es = atomic_data[:, 1] * PERCMTOEV / RY  # Ry
    F2 = atomic_data[:, 4] * 1e13  # 1/s

    if label == "he":
        g0 = 1

    elif label == "li":
        g0 = 2
        Esli = 1 / (atomic_data[:, 0] * 1e-8) - atomic_data[:, 1]
        # Difference between energy levels
        Es = Esli * PERCMTOEV / RY
    else:
        raise ValueError(f"wrong label given: {label}")
    # fmt: off
    _Te = Te[:, None]
    _F2 = F2[None, ]
    _Es = Es[None, ]
    intensity = (1 / g0 * 4 * np.pi ** (3 / 2) * a0 ** 3 / _Te ** (3 / 2) * _F2
                 * np.exp(-_Es / _Te))
    # fmt: on
    return intensity


class MARCHUKReader:
    """
    Class for interacting with Marchuks data format and return PECs in format:
    Dataset of dims(line name, electron temperature) and data variables of emission type
    """

    def __init__(
        self,
        extrapolate: bool = True,
        filehead: str = None,
        element: str = "ar",
        charge: int = 16,
    ):
        if filehead is None:
            filehead = FILEHEAD
        self.filehead = filehead
        self.extrapolate = extrapolate
        self.element = element
        self.charge = charge

        self.pec_rawdata = self.build_pec_database()
        self.pecs = self._make_pecs_dataarray()

    def build_pec_database(
        self,
        Te: np.typing.ArrayLike = np.linspace(200, 10000, 10000),
    ):
        """
        Reads Marchuk's Atomic data and builds DataArrays for each emission type
        with co-ordinates of line label and electron temperature
        Input is the Te vector for dielectronic recombination data
        """

        lines_main = ["W", "X", "Y", "Z"]
        lines_ise = ["q", "r", "s", "t", "u", "v", "w"]
        lines_isi = ["Z"]
        lines_casc = ["q", "r", "s", "t"]

        # Wavelengths from the paper from O. Marchuk
        # "Modelling of helium like spectra at TEXTOR and TORE SUPRA"
        w0 = 0.39492
        x0 = 0.39660
        y0 = 0.39695
        z0 = 0.39943
        q0 = 0.39815
        r0 = 0.39836
        s0 = 0.39678
        t0 = 0.39687
        u0 = 1  # Ignore
        v0 = 1  # Ignore

        wavelengths_main = np.array([w0, x0, y0, z0])
        wavelengths_ise = np.array([q0, r0, s0, t0, u0, v0, v0])
        wavelengths_isi = np.array([z0])
        wavelengths_casc = np.array([q0, r0, s0, t0])

        # Read in atomic data
        # Excitation / Recombination / Charge Exchange /
        # Inner-Shell Excitation / Inner-Shell Ionisation / Cascades
        # Missing ion - ion excitation (not relevant till ne>>1E20 1/m3)

        # exc = np.loadtxt(head + "DirectRates.dat", skiprows=1)
        conversion_factor = 1e-6  # cm^3 -> m^3
        exc = np.loadtxt(self.filehead + "WXYZ_R-matrix.txt", comments="#")
        recom = np.loadtxt(self.filehead + "RecombRates.dat", skiprows=1)
        cxr = np.loadtxt(self.filehead + "ChargeRates.dat", skiprows=1)
        ise = np.loadtxt(self.filehead + "LiCollSatt.dat", skiprows=1)
        isi = np.loadtxt(self.filehead + "InnerRates.dat", skiprows=1)
        casc = np.loadtxt(self.filehead + "Cascade.dat", skiprows=5)

        # Dielectronic recombination / wavelength; Es; Ar; Aa; F2; Satellites
        n2 = np.loadtxt(
            self.filehead + "n2dielsat.dat", skiprows=1, usecols=(0, 1, 2, 3, 4, 5)
        )
        n3 = np.loadtxt(
            self.filehead + "n3dielsat.dat", skiprows=1, usecols=(0, 1, 2, 3, 4, 5)
        )
        n4 = np.loadtxt(
            self.filehead + "n4dielsat.dat", skiprows=1, usecols=(0, 1, 2, 3, 4, 5)
        )
        n5 = np.loadtxt(
            self.filehead + "n5dielsat.dat", skiprows=1, usecols=(0, 1, 2, 3, 4, 5)
        )
        lin2 = np.loadtxt(
            self.filehead + "n2lidielsat.dat", skiprows=1, usecols=(0, 1, 2, 3, 4, 5)
        )
        # Use line labels from file
        lines_n2 = np.genfromtxt(
            self.filehead + "n2dielsat.dat", skip_header=1, usecols=(6), dtype="str"
        )
        lines_n3 = np.genfromtxt(
            self.filehead + "n3dielsat.dat", skip_header=1, usecols=(6), dtype="str"
        )
        lines_n4 = np.genfromtxt(
            self.filehead + "n4dielsat.dat", skip_header=1, usecols=(6), dtype="str"
        )
        lines_n5 = np.genfromtxt(
            self.filehead + "n5dielsat.dat", skip_header=1, usecols=(6), dtype="str"
        )
        lines_lin2 = np.genfromtxt(
            self.filehead + "n2lidielsat.dat", skip_header=1, usecols=(6), dtype="str"
        )

        # replace duplicate line "name" with "name"+ str(n)
        line_names = lines_n2.tolist()
        lines_n2 = [
            v + str(line_names[:i].count(v) + 1) if line_names.count(v) > 1 else v
            for i, v in enumerate(line_names)
        ]
        line_names = lines_n3.tolist()
        lines_n3 = [
            v + str(line_names[:i].count(v) + 1) if line_names.count(v) > 1 else v
            for i, v in enumerate(line_names)
        ]
        line_names = lines_n4.tolist()
        lines_n4 = [
            v + str(line_names[:i].count(v) + 1) if line_names.count(v) > 1 else v
            for i, v in enumerate(line_names)
        ]
        line_names = lines_n5.tolist()
        lines_n5 = [
            v + str(line_names[:i].count(v) + 1) if line_names.count(v) > 1 else v
            for i, v in enumerate(line_names)
        ]
        line_names = lines_lin2.tolist()
        lines_lin2 = [
            v + str(line_names[:i].count(v) + 1) if line_names.count(v) > 1 else v
            for i, v in enumerate(line_names)
        ]

        rates_n2 = diel_calc(n2, Te)
        rates_n3 = diel_calc(n3, Te)
        rates_n4 = diel_calc(n4, Te)
        rates_n5 = diel_calc(n5, Te)
        rates_lin2 = diel_calc(lin2, Te, label="li")

        # Convert data to DataArrays
        exc_array = xr.DataArray(
            data=exc[:, 1:, np.newaxis] * conversion_factor,
            coords={
                "electron_temperature": exc[:, 0] * 1e3,
                "line_name": lines_main,
                "type": ["excit"],
                "wavelength": (
                    ("line_name",),
                    wavelengths_main,
                ),
            },
            dims=["electron_temperature", "line_name", "type"],
        )
        recom_array = xr.DataArray(
            data=recom[:, 1:, np.newaxis] * conversion_factor,
            coords={
                "electron_temperature": recom[:, 0] * 1e3,
                "line_name": lines_main,
                "type": ["recom"],
                "wavelength": (
                    "line_name",
                    wavelengths_main,
                ),
            },
            dims=["electron_temperature", "line_name", "type"],
        )
        cxr_array = xr.DataArray(
            data=cxr[:, 1:5, np.newaxis] * conversion_factor,
            coords={
                "electron_temperature": cxr[:, 0] * 1e3,
                "line_name": lines_main,
                "type": ["cxr"],
                "wavelength": ("line_name", wavelengths_main),
            },
            dims=["electron_temperature", "line_name", "type"],
        )

        ise_array = xr.DataArray(
            data=ise[:, 1:, np.newaxis] * conversion_factor,
            coords={
                "electron_temperature": ise[:, 0] * 1e3,
                "line_name": lines_ise,
                "type": ["ise"],
                "wavelength": ("line_name", wavelengths_ise),
            },
            dims=["electron_temperature", "line_name", "type"],
        )
        isi_array = xr.DataArray(
            data=isi[:, 1:, np.newaxis] * conversion_factor,
            coords={
                "electron_temperature": isi[:, 0] * 1e3,
                "line_name": lines_isi,
                "type": ["isi"],
                "wavelength": ("line_name", wavelengths_isi),
            },
            dims=["electron_temperature", "line_name", "type"],
        )

        casc_factor_array = xr.DataArray(
            data=casc[:, 1:, np.newaxis],
            coords={
                "electron_temperature": casc[
                    :,
                    0,
                ]
                * 1e3,
                "line_name": lines_casc,
                "type": ["diel"],
                "wavelength": ("line_name", wavelengths_casc),
            },
            dims=["electron_temperature", "line_name", "type"],
        )

        n2_array = xr.DataArray(
            data=rates_n2[:, :, np.newaxis] * conversion_factor,
            coords={
                "electron_temperature": Te,
                "line_name": lines_n2,
                "type": ["diel"],
                "wavelength": ("line_name", n2[:, 0] * 0.1),
            },
            dims=["electron_temperature", "line_name", "type"],
        )

        n3_array = xr.DataArray(
            data=rates_n3[:, :, np.newaxis] * conversion_factor,
            coords={
                "electron_temperature": Te,
                "line_name": lines_n3,
                "type": ["diel"],
                "wavelength": ("line_name", n3[:, 0] * 0.1),
            },
            dims=["electron_temperature", "line_name", "type"],
        )

        n4_array = xr.DataArray(
            data=rates_n4[:, :, np.newaxis] * conversion_factor,
            coords={
                "electron_temperature": Te,
                "line_name": lines_n4,
                "type": ["diel"],
                "wavelength": ("line_name", n4[:, 0] * 0.1),
            },
            dims=["electron_temperature", "line_name", "type"],
        )

        n5_array = xr.DataArray(
            data=rates_n5[:, :, np.newaxis] * conversion_factor,
            coords={
                "electron_temperature": Te,
                "line_name": lines_n5,
                "type": ["diel"],
                "wavelength": ("line_name", n5[:, 0] * 0.1),
            },
            dims=["electron_temperature", "line_name", "type"],
        )

        lin2_array = xr.DataArray(
            data=rates_lin2[:, :, np.newaxis] * conversion_factor,
            coords={
                "electron_temperature": Te,
                "line_name": lines_lin2,
                "type": ["li_diel"],
                "wavelength": ("line_name", lin2[:, 0] * 0.1),
            },
            dims=["electron_temperature", "line_name", "type"],
        )

        # Cascade functions
        casc_factor = casc_factor_array.interp(
            electron_temperature=Te, method="quadratic"
        )

        q_idx = n2_array.line_name.str.contains("q!").values
        r_idx = n2_array.line_name.str.contains("r!").values
        s_idx = n2_array.line_name.str.contains("s!").values
        t_idx = n2_array.line_name.str.contains("t!").values
        casc_idx = q_idx | r_idx | s_idx | t_idx
        n2_array.loc[dict(line_name=casc_idx)] = n2_array.sel(line_name=casc_idx) * (
            1 + casc_factor.values
        )

        # Atomic data
        pec_database = dict(
            excit=exc_array,
            recom=recom_array,
            cxr=cxr_array,
            ise=ise_array,
            isi=isi_array,
            n2=n2_array,
            n3=n3_array,
            n4=n4_array,
            n5=n5_array,
            li_n2=lin2_array,
        )
        return pec_database

    def _interp_pecs(self, Te: np.typing.ArrayLike = np.linspace(200, 10000, 10000)):
        _interp_pec = {}
        for _pec_name, _pec in self.pec_rawdata.items():
            _interp_pec[_pec_name] = _pec.interp(
                electron_temperature=Te, kwargs={"fill_value": "extrapolate"}
            )
        return _interp_pec

    def _make_pecs_dataarray(self):
        """
        TODO: add caching of dataset
        Returns
        -------
        """
        _pecs = self._interp_pecs()
        _dataset = {}
        for _pec_name, _pec in _pecs.items():
            _dataset[_pec_name] = _pec.to_dataset(dim="type")
        _dataset = xr.merge([*_dataset.values()])
        dataarray = _dataset.to_array(dim="type")
        dataarray = dataarray.transpose(*["electron_temperature", "line_name", "type"])
        return dataarray



"""
IndicA is a A library for **In**tegrated **Di**agnosti**C** **A**nalysis
of magnetic-confined fusion devices
"""

from .abstractio import BaseIO
from .data import InDiCAArrayAccessor
from .data import InDiCADatasetAccessor
from .equilibrium import Equilibrium
from .plasma import Plasma
from .plasma import PlasmaProfiler

__all__ = [
    "InDiCAArrayAccessor",
    "InDiCADatasetAccessor",
    "BaseIO",
    "Equilibrium",
    "Plasma",
    "PlasmaProfiler",
]



"""Provides implementation of :py:class:`readers.DataReader` for reading PPF data
produced by JET

"""

from pathlib import Path
from typing import Any
from typing import Dict
from typing import Tuple

import numpy as np

from indica.abstractio import BaseIO
from indica.configs.readers.jetconf import JETConf
from indica.configs.readers.machineconf import MachineConf
from indica.converters import CoordinateTransform
from indica.converters import LineOfSightTransform
from indica.converters import TransectCoordinates
from indica.converters import TrivialTransform
from indica.numpy_typing import ArrayLike
from indica.readers.datareader import DataReader
from indica.readers.salutils import SALUtils
from indica.readers.surfutils import read_surf_los


class JETReader(DataReader):
    """Class to read JET PPF data using SAL"""

    def __init__(
        self,
        pulse: int,
        tstart: float,
        tend: float,
        machine_conf: MachineConf = JETConf,
        reader_utils: BaseIO = SALUtils,
        server: str = "https://sal.jet.uk",
        verbose: bool = False,
        default_error: float = 0.05,
        *args,
        **kwargs,
    ):
        super().__init__(
            pulse,
            tstart,
            tend,
            machine_conf=machine_conf,
            reader_utils=reader_utils,
            server=server,
            verbose=verbose,
            default_error=default_error,
            **kwargs,
        )
        self.reader_utils = self.reader_utils(pulse, server)

    def _get_equilibrium(
        self,
        data: dict,
    ) -> Tuple[Dict[str, Any], CoordinateTransform]:
        data["t"] = data["rbnd_dimensions"][0]
        data["index"] = data["rbnd_dimensions"][1]
        data["psin"] = data["f_dimensions"][1]
        uid = data["uid"]
        instrument = data["instrument"]
        revision = data["revision"]
        data["psi_r"], data["psi_r_records"] = self.reader_utils.get_signal(
            uid, instrument, "psir", revision
        )
        data["psi_z"], data["psi_z_records"] = self.reader_utils.get_signal(
            uid, instrument, "psiz", revision
        )
        data["psi"] = data["psi"].reshape(
            (len(data["t"]), len(data["psi_z"]), len(data["psi_r"]))
        )
        data["psi_error"] = data["psi_error"].reshape(
            (len(data["t"]), len(data["psi_z"]), len(data["psi_r"]))
        )
        transform = assign_trivial_transform()
        return data, transform

    def _get_thomson_scattering(
        self,
        data: dict,
    ) -> Tuple[Dict[str, Any], CoordinateTransform]:
        data["R"] = data["z_dimensions"][0]
        data["x"] = data["R"]
        data["y"] = np.zeros_like(data["R"])
        data["t"] = data["te_dimensions"][0]
        data["channel"] = np.arange(len(data["R"]))
        transform = assign_transect_transform(data)
        return data, transform

    def _get_cyclotron_emissions(
        self,
        data: dict,
    ) -> Tuple[Dict[str, Any], CoordinateTransform]:
        raise NotImplementedError

    def _get_density_reflectometer(
        self,
        data: dict,
    ) -> Tuple[Dict[str, Any], CoordinateTransform]:
        raise NotImplementedError

    def _get_charge_exchange(
        self,
        data: dict,
    ) -> Tuple[Dict[str, Any], CoordinateTransform]:
        data["R"] = data["R"].mean(0)
        data["x"] = data["R"]
        data["y"] = np.zeros_like(data["R"])
        data["z"] = data["z"].mean(0)
        data["t"] = data["R_dimensions"][0]
        data["channel"] = np.arange(len(data["R"]))
        transform = assign_transect_transform(data)
        return data, transform

    def _get_radiation(
        self,
        data: dict,
    ) -> Tuple[Dict[str, Any], CoordinateTransform]:
        uid = data["uid"]
        instrument = data["instrument"]
        revision = data["revision"]
        if "sxr" in instrument.lower():
            quantity = instrument[-1]
            _instrument = "sxr"
            luminosities = []
            channels = []
            for i in range(
                1, self.machine_conf._RADIATION_RANGES[_instrument + "/" + quantity] + 1
            ):
                try:
                    qval, q_dims, _, q_path = self.reader_utils.get_data(
                        uid, _instrument, f"{quantity}{i:02d}", revision
                    )
                    luminosities.append(qval)
                    channels.append(i)
                except Exception:
                    continue
                if data.get("t") is None:
                    data["t"] = q_dims[0]
            data["brightness"] = np.array(luminosities).T
            data["channel"] = channels
        elif "kb5" in instrument.lower():
            _instrument = "bolo"
            quantity = instrument
            qval, qval_dimensions, _, qval_records = self.reader_utils.get_data(
                uid=uid, instrument="bolo", quantity=quantity, revision=revision
            )
            data["brightness"] = qval
            data["t"] = qval_dimensions[0]
            data["brightness_records"] = qval_records
            data["channel"] = np.arange(len(qval_dimensions[1]))
        else:
            raise UserWarning(f"{instrument} unsupported for {__class__}")

        xstart, xend, zstart, zend, ystart, yend = read_surf_los(
            self.machine_conf.SURF_PATH,
            self.pulse,
            _instrument.lower() + "/" + quantity.lower(),
        )
        location = np.asarray([xstart, ystart, zstart])
        direction = np.asarray([xend, yend, zend]) - location
        data["location"] = location.transpose()
        data["direction"] = direction.transpose()
        transform = assign_lineofsight_transform(data)
        return data, transform

    def _get_zeff(
        self,
        data: dict,
    ) -> Tuple[Dict[str, Any], CoordinateTransform]:
        uid = data["uid"]
        instrument = data["instrument"]
        revision = data["revision"]
        quantity = instrument[-1]
        instrument = instrument[:-1]
        qval, qval_dimensions, _, qval_path = self.reader_utils.get_data(
            uid=uid, instrument="ks3", quantity="zef" + quantity, revision=revision
        )
        data["zeff_avrg"] = qval
        data["t"] = qval_dimensions[0]
        data["zeff_avrg_records"] = qval_path
        los, los_path = self.reader_utils.get_signal(
            uid,
            self.machine_conf._BREMSSTRAHLUNG_LOS[instrument],
            "los" + quantity,
            revision,
        )
        data["location"] = np.asarray([[(los[1] / 1000), 0, (los[2] / 1000)]])
        data["direction"] = (
            np.asarray([[(los[4] / 1000), 0, (los[5] / 1000)]]) - data["location"]
        )
        transform = assign_lineofsight_transform(data)
        return data, transform


def assign_lineofsight_transform(database_results: Dict):
    transform = LineOfSightTransform(
        database_results["location"][:, 0],
        database_results["location"][:, 1],
        database_results["location"][:, 2],
        database_results["direction"][:, 0],
        database_results["direction"][:, 1],
        database_results["direction"][:, 2],
        machine_dimensions=database_results["machine_dims"],
        dl=database_results["dl"],
        passes=database_results["passes"],
    )
    return transform


def assign_transect_transform(database_results: Dict):
    transform = TransectCoordinates(
        database_results["x"],
        database_results["y"],
        database_results["z"],
        machine_dimensions=database_results["machine_dims"],
    )

    return transform


def assign_trivial_transform():
    transform = TrivialTransform()
    return transform


def _get_cxrs_los_geometry(sav_file: Path, tracks: ArrayLike) -> Any:
    """Read IDL save file to get position and direction for KS5 tracks

    Parameters
    ----------

    sav_file:
        Path to IDL save file to load
    tracks:
        ArrayLike of tracks to select from save file, from JET dtype `TRCK`

    Returns
    -------

    :
        Tuple of position and direction arrays
    """
    import scipy.io as io

    data = io.readsav(sav_file)
    fibres = data.ptrfib.fibres[0]
    numfibview = fibres.numfibview[0].astype(str)
    origin, direction = [], []
    for name, pos, vec in zip(
        numfibview,
        fibres.losdef[0].virtualposition_roomtemp_rot[0].cartesian_ref[0],
        fibres.losdef[0].virtualdirection_rot[0].cartesian_ref[0],
    ):
        if name in tracks:
            origin.append(pos / 1000)  # mm -> m
            direction.append(vec)
    return (np.asarray(origin), np.asarray(direction))


def _setup_idl(pulse: int) -> Any:
    import idlbridge as idlb

    idlb.execute(".reset")
    idlb.execute(
        "!PATH=!PATH + ':' + "
        "expand_path( '+~cxs/idl_spectro/' ) + ':' + "
        "expand_path( '+~cxs/idl_spectro/show' ) + ':' + "
        "expand_path( '+~cxs/ks6read/' ) + ':' + "
        "expand_path( '+~cxs/ktread/' ) + ':' + "
        "expand_path( '+~cxs/kx1read/' ) + ':' + "
        "expand_path( '+~cxs/idl_spectro/kt3d' ) + ':' + "
        "expand_path( '+~cxs/utc' ) + ':' + "
        "expand_path( '+~cxs/instrument_data' ) + ':' + "
        "expand_path( '+~cxs/calibration' ) + ':' + "
        "expand_path( '+~cxs/alignment' ) + ':' + "
        "expand_path( '+/usr/local/idl' ) + ':' + "
        "expand_path( '+/home/CXSE/cxsfit/idl/' ) + ':' + "
        "expand_path( '+~/jet/share/lib' ) + ':' + "
        "expand_path( '+~/jet/share/root/lib' ) + ':' + "
        "expand_path( '+~/jet/share/idl' )"
    )

    idlb.execute(
        "!PATH = !PATH + ':' + expand_path('+/u/cxs/utilities',/all_dirs)+ ':'"
    )

    idlb.execute(
        "!PATH = !PATH + ':' + expand_path('+/u/cxs/instrument_data/namelists')+ ':' + "
        "expand_path('+/u/cxs/instrument_data/jpfnodes')"
    )

    idlb.execute("!PATH = !PATH + ':' + expand_path('+/usr/local/idl')")
    idlb.execute("!PATH = !PATH + ':' + expand_path('+/home/CXSE/cxsfit/idl/')")
    idlb.execute("!PATH = !PATH +':/home/CXSE/cxsfit/idl:'")
    idlb.execute(".compile plot")
    idlb.execute(".compile ppfread")
    idlb.execute(".compile cxf_number_to_text")
    idlb.execute(".compile cxf_read_switches")
    idlb.execute(".compile cxf_decompress_history")

    idlb.put("shot", pulse)
    idlb.execute("julian_date = agm_pulse_to_julian(shot, 'DG')")

    return idlb


def _get_cxrs_los_savfile(pulse: int, spec: str) -> Path:
    """Determine correct sav file for given pulse and spectrometer

    Parameters
    ----------
    pulse:
        Pulse to search for
    spec:
        Spectrometer to search for

    Returns
    -------
    :
        Path to save file
    """
    octant = {"ks5a": 7, "ks5b": 1, "ks5c": 7, "ks5d": 7, "ks5e": 1}.get(spec.lower())
    idlb = _setup_idl(pulse=pulse)
    idlb.execute(
        "str1 = periscope_oct{}_hist_align(julian_date=julian_date)".format(str(octant))
    )
    idlb.execute("file_align = str1.file_align")
    savfile = Path(str(idlb.get("file_align")))
    assert savfile.exists() and savfile.is_file()
    return savfile


def _get_cxrs_active_tracks(pulse: int, spec: str, trck: ArrayLike) -> ArrayLike:
    """Translate tracks used to track names as in geometry save file"""
    idlb = _setup_idl(pulse=pulse)
    idlb.execute(
        "str1={}_hist_fibresetup(pulse=pulse,julian_date=julian_date)".format(str(spec))
    )
    idlb.execute("viewing_position=str1.pulse_setup.viewing_position")
    idlb.execute("track_reshuffle=str1.pulse_setup.ks4fit_track_reshuffle")
    viewing_position = idlb.get("viewing_position")
    track_reshuffle = idlb.get("track_reshuffle")
    assert len(viewing_position) == len(track_reshuffle) == len(trck)
    return [
        viewing_position[i - 1]
        for i in track_reshuffle
        if int(trck[len(trck) - i]) == 1
    ][::-1]



from typing import List
from typing import Tuple

from MDSplus import Connection
import numpy as np

from indica import BaseIO
from ..numpy_typing import RevisionLike


class MDSError(Exception):
    """An exception which occurs when trying to read MDS+ data which would
    not be caught by the lower-level MDSplus library. An example would be
    failing to find any valid channels for an instrument when each channel
    is a separate DTYPE.

    """


class MDSWarning(UserWarning):
    """A warning that occurs while trying to read MDS+ data. Typically
    related to caching in some way.

    """


# this will be baseio class instead. what is defauly pulse?
class MDSUtils(BaseIO):
    def __init__(
        self,
        pulse,
        server: str = "smaug",
        tree: str = "ST40",
    ):
        self.tree: str = tree
        self.pulse: int = pulse
        self.conn: Connection = Connection(server)
        self.conn.openTree(self.tree, self.pulse)

    def close(self) -> None:
        del self.conn

    @property
    def requires_authentication(self) -> bool:
        return False

    def get_signal(
        self, uid: str, instrument: str, quantity: str, revision: RevisionLike
    ) -> Tuple[np.array, str]:
        """Gets the signal for the given INSTRUMENT, at the
        given revision."""
        path, path_check = self.get_mds_path(uid, instrument, quantity, revision)
        if quantity.lower() == ":best_run":
            data = str(self.conn.get(path))
        else:
            data = np.array(self.conn.get(path))

        return data, path

    def get_signal_dims(
        self,
        mds_path: str,
        ndims: int,
    ) -> Tuple[List[np.array], List[str]]:
        """Gets the dimensions of a signal given the path to the signal
        and the number of dimensions"""

        dimensions = []
        paths = []
        for dim in range(ndims):
            path = f"dim_of({mds_path},{dim})"
            dim_tmp = self.conn.get(path).data()

            paths.append(path)
            dimensions.append(np.array(dim_tmp))
        return dimensions, paths

    def get_signal_units(
        self,
        mds_path: str,
    ) -> str:
        """Gets the units of a signal given the path to the signal
        and the number of dimensions"""

        path = f"units_of({mds_path})"
        unit = self.conn.get(path).data()

        return unit

    def get_data(
        self, uid: str, instrument: str, quantity: str, revision: RevisionLike
    ) -> Tuple[np.array, List[np.array], str, str]:
        """Gets the signal and its coordinates for the given INSTRUMENT, at the
        given revision."""
        data, _path = self.get_signal(uid, instrument, quantity, revision)
        dims, _ = self.get_signal_dims(_path, len(data.shape))
        unit = self.get_signal_units(_path)

        return data, dims, unit, _path

    def revision_name(self, revision: RevisionLike) -> str:
        """Return string defining RUN## or BEST if revision = 0"""

        if type(revision) == int:
            _revision = int(revision)
            if _revision < 0:
                rev_str = ""
            elif _revision == 0:
                rev_str = "best"
            elif _revision < 10:
                rev_str = f"run0{int(_revision)}"
            else:
                rev_str = f"run{int(_revision)}"
        else:
            rev_str = f"{revision}"

        return rev_str.upper()

    def get_best_revision(self, uid: str, instrument: str):
        """
        Return revision name to which BEST is pointing to
        """
        best_revision, _ = self.get_signal(uid, instrument, ".best_run", "best")
        return best_revision

    def get_revision(self, uid: str, instrument: str, revision: RevisionLike) -> str:
        """
        Return revision name given
        """
        revision_name = self.revision_name(revision)
        if revision_name == "BEST":
            revision_name = self.get_best_revision(uid, instrument)

        return revision_name

    def get_mds_path(
        self, uid: str, instrument: str, quantity: str, revision: RevisionLike
    ) -> Tuple[str, str]:
        """Return the path in the MDS+ database to for the given INSTRUMENT/CODE

        uid: currently redundant --> set to empty string ""
        instrument: e.g. "efit"
        quantity: e.g. ".global:cr0" # minor radius
        revision: if 0 --> looks for "best", else "run##"
        """
        revision_name = self.revision_name(revision)
        mds_path = ""
        if len(uid) > 0:
            mds_path += f".{uid}".upper()
        if len(instrument) > 0 and instrument.upper() != self.tree.upper():
            mds_path += f".{instrument}".upper()
        mds_path += f".{revision_name}{quantity}".upper()
        return mds_path, self.mdsCheck(mds_path)

    def mdsCheck(self, mds_path):
        """Return FAILED if node doesn't exist or other error
        Return FAILED if: lenght(data)==1 and data==nan"""

        mds_path_test = (
            f"_dummy = IF_ERROR (IF ((SIZE ({mds_path})==1), "
            + f'IF ({mds_path}+1>{mds_path}, {mds_path}, "FAILED"),'
            + f' {mds_path}), "FAILED")'
        )

        return mds_path_test



from copy import deepcopy

import matplotlib.pylab as plt
import numpy as np
import pandas as pd
import xarray as xr

from indica import Equilibrium
from indica.readers.read_st40 import bin_data_in_time
from indica.readers.read_st40 import ReadST40
from indica.utilities import FIG_PATH
from indica.utilities import save_figure
from indica.utilities import set_axis_sci
from indica.utilities import set_plot_colors
from indica.utilities import set_plot_rcparams
import indica.workflows.load_modelling_plasma as load_modelling
from indica.workflows.load_modelling_plasma import initialize_diagnostic_models
from indica.workflows.load_modelling_plasma import plasma_code

plt.ion()

CMAP, COLORS = set_plot_colors()


def read_modelling_runs(
    tstart: float = 0.0,
    tend: float = 0.2,
    dt: float = 0.01,
    pulse: int = 13110009,
    code: str = "astra",
    runs: list = None,
):

    if runs is None:
        runs = [f"RUN{run}" for run in (500 + np.arange(61, 77))]

    print(f"Reading {code} data")
    code_raw_data: dict = {}
    code_binned_data: dict = {}
    # equilibrium: dict = {}
    code_reader = ReadST40(pulse, tstart, tend, dt=dt, tree=code)
    for run in runs:
        code_raw_data[run] = code_reader.get_raw_data("", code, run)
        code_binned_data[run] = bin_data_in_time(code_reader.raw_data, tstart, tend, dt)

    return code_raw_data, code_binned_data


def plot_iaea_tm_2023(data=None, savefig=False, ext="png"):

    pulses = [9539, 9783, 10009]

    data = compare_pulses(
        pulses,
        data=data,
        qpop=[
            "mhd:ampl_odd_n",
            "mag:vloop",
            "diode_detr:filter_001",
            "xrcs:te_kw",
            "xrcs:ti_w",
            "cxrs_pi:ti",
            "cxrs_pi:vtor",
        ],
        savefig=savefig,
        ext=ext,
    )

    #
    # data = compare_pulses(
    #     pulses,
    #     data=data,
    #     qpop=[
    #         # "diode_detr:filter_001",
    #         "mhd:ampl_odd_n",
    #         "mag:vloop",
    #         "efit:ipla",
    #         "efit:wp",
    #         "smmh1:ne_bar",
    #         "lines:h_alpha",
    #         "nbi:pin",
    #     ],
    #     figname="kinetics",
    #     savefig=savefig,
    #     ext=ext,
    # )
    #
    return data


def plot_stan_ppcf_time_evol(data=None, savefig=False, ext="png"):

    pulses = [9520, 9539, 9780, 10009]

    data = compare_pulses(
        pulses,
        data=data,
        qpop=[
            "efit:ipla",
            "efit:wp",
            "nbi:pin",
            "mag:vloop",
            "smmh1:ne_bar",
            "smmh1:ne",
            "nirh1:ne",
            "halpha:brightness",
            "brems:brightness",
            "mhd:ampl_odd_n",
        ],
        savefig=savefig,
        ext=ext,
        add_pulse_label_to="cxff_pi:vtor",  # "sxr_diode_1:brightness",
    )

    return data


def build_astra_profiles(
    pulse: int = 9520,
    runs: list = None,
    plot: bool = True,
    save_fig: bool = False,
    write_csv: bool = False,
):
    tstart = 0.02
    tend = 0.12
    dt = 0.01
    runs_all = {
        9520: ["RUN561", "RUN562", "RUN569", "RUN570"],
        9539: ["RUN562", "RUN564", "RUN569", "RUN570", "RUN572"],
        9780: ["RUN564", "RUN566", "RUN572", "RUN573", "RUN574"],
        10009: ["RUN560", "RUN564", "RUN565", "RUN573"],
    }
    code = "astra"
    equil = "efit"  # code
    verbose = False
    instruments = ["xrcs", "cxff_pi", "smmh1", "nirh1", "efit"]

    pulse_code = int(13.1e6 + pulse)

    if pulse in runs_all.keys():
        runs = runs_all[pulse]
    else:
        raise ValueError("Pulse not in runs dictionary")

    code_raw_data, code_binned_data, tstart, tend = load_modelling.read_modelling_runs(
        code,
        pulse_code,
        runs,
        tstart=tstart,
        tend=tend,
        dt=dt,
    )
    runs = list(code_raw_data.keys())

    if pulse is not None:
        print("Reading ST40 data")
        st40 = ReadST40(pulse, tstart, tend, dt=dt, tree="st40")
        st40(instruments=instruments, map_diagnostics=False)

    if equil != code:
        equilibrium = {run: Equilibrium(st40.raw_data[equil]) for run in runs}
    else:
        equilibrium = {run: Equilibrium(code_raw_data[run]) for run in runs}

    plasma = plasma_code(
        pulse_code,
        tstart,
        tend,
        dt,
        code_raw_data,
        verbose=verbose,
        equilibrium=equilibrium,
    )

    print("Initializing diagnostic models")
    models_to_run = initialize_diagnostic_models(st40.binned_data)

    models: dict = {}
    bckc: dict = {}
    print("Running diagnostic models")
    for run in runs:
        print(run)
        bckc[run] = {}
        models[run] = {}
        _plasma = plasma[run]
        for instrument in models_to_run.keys():
            if hasattr(models_to_run[instrument], "los_transform"):
                models_to_run[instrument].los_transform.set_equilibrium(
                    equilibrium[run], force=True
                )
            if hasattr(models_to_run[instrument], "transect_transform"):
                models_to_run[instrument].transect_transform.set_equilibrium(
                    equilibrium[run], force=True
                )

            models_to_run[instrument].set_plasma(_plasma)

            _bckc = models_to_run[instrument](moment_analysis=True)
            bckc[run][instrument] = deepcopy(_bckc)
            # models[run][instrument] = deepcopy(models_to_run[instrument])

    if plot:
        plot_stan_ppcf_profiles(
            st40, bckc, plasma, save_fig=save_fig, write_csv=write_csv
        )

    return st40, bckc, models_to_run, plasma


def plot_stan_ppcf_profiles(
    st40=None,
    bckc=None,
    plasma=None,
    tplot: float = None,
    ylim: tuple = None,
    save_fig: bool = False,
    write_csv: bool = False,
):
    if st40 is None or bckc is None or plasma is None:
        st40, bckc, models_to_run, plasma = build_astra_profiles()

    CMAP, COLORS = set_plot_colors()
    set_plot_rcparams("profiles")
    cxff_pi_chans_all = {
        9520: slice(2, 3),
        9539: slice(2, 3),
        9780: slice(2, 3),
        10009: slice(2, 3),
    }
    R_shift_all = {
        9520: [-0.01, 0.01],
        9539: [-0.02, 0.0],
        9780: [0.0, 0.02],
        10009: [0.00, 0.02],
    }
    tplot_all = {9520: 0.09, 9539: 0.071, 9780: 0.081, 10009: 0.059}
    te_key = {9520: "mean", 9539: "mean", 9780: "n3w", 10009: "mean"}
    if tplot is None and st40.pulse in tplot_all.keys():
        tplot = tplot_all[st40.pulse]
        R_shift_scan = R_shift_all[st40.pulse]
        cxff_pi_chans = cxff_pi_chans_all[st40.pulse]
    if tplot is None:
        raise ValueError("Enter valid tplot")

    __Ne = []
    __Te = []
    __Ti = []
    __Ti_w_pos = []
    __Ti_w_pos_err_in = []
    __Ti_w_pos_err_out = []
    runs = list(bckc)
    for run in runs:
        t = plasma[run].t.sel(t=tplot, method="nearest")
        __Ne.append(plasma[run].electron_density.sel(t=t))
        __Te.append(plasma[run].electron_temperature.sel(t=t))
        __Ti.append(
            plasma[run].ion_temperature.sel(element=plasma[run].main_ion).sel(t=t)
        )
        __Ti_w_pos.append(bckc[run]["xrcs"]["ti_w"].pos.sel(t=t))
        __Ti_w_pos_err_in.append(bckc[run]["xrcs"]["ti_w"].pos_err_in.sel(t=t))
        __Ti_w_pos_err_out.append(bckc[run]["xrcs"]["ti_w"].pos_err_out.sel(t=t))

    _Ne = xr.concat(__Ne, "run").assign_coords(run=runs)
    _Te = xr.concat(__Te, "run").assign_coords(run=runs)
    _Ti = xr.concat(__Ti, "run").assign_coords(run=runs)
    _Ti_w_pos = xr.concat(__Ti_w_pos, "run").assign_coords(run=runs)
    _Ti_w_pos_err_in = xr.concat(__Ti_w_pos_err_in, "run").assign_coords(run=runs)
    _Ti_w_pos_err_out = xr.concat(__Ti_w_pos_err_out, "run").assign_coords(run=runs)

    Ne = _Ne.mean("run")
    Te = _Te.mean("run")
    Ti = _Ti.mean("run")
    Ti_w = st40.raw_data["xrcs"]["ti_w"].sel(t=tplot, method="nearest")
    Ti_w_err = st40.raw_data["xrcs"]["ti_w"].error.sel(t=tplot, method="nearest")
    Te_kw = st40.raw_data["xrcs"]["te_kw"].sel(t=tplot, method="nearest")
    Te_kw_err = st40.raw_data["xrcs"]["te_kw"].error.sel(t=tplot, method="nearest")
    Te_n3w = st40.raw_data["xrcs"]["te_n3w"].sel(t=tplot, method="nearest")
    Te_n3w_err = st40.raw_data["xrcs"]["te_n3w"].error.sel(t=tplot, method="nearest")
    Ti_cxff_pi = (
        st40.raw_data["cxff_pi"]["ti"]
        .sel(t=tplot, method="nearest")
        .sel(channel=cxff_pi_chans)
    )
    Ti_cxff_pi_err = (
        st40.raw_data["cxff_pi"]["ti"]
        .error.sel(t=tplot, method="nearest")
        .sel(channel=cxff_pi_chans)
    )

    _Ti_cxff_pi_pos_all = []
    for R_shift in R_shift_scan:
        R = Ti_cxff_pi.R + R_shift
        z = Ti_cxff_pi.z * 0
        pos, _, _ = plasma[run].equilibrium.flux_coords(R, z, Ti_cxff_pi.t)
        _Ti_cxff_pi_pos_all.append(pos)

    Ti_cxff_pi_pos_all = xr.concat(_Ti_cxff_pi_pos_all, "R_shift").assign_coords(
        R_shift=R_shift_scan
    )

    Te_mean = np.mean([Te_kw, Te_n3w])
    err_kw = (Te_kw_err / Te_kw) ** 2
    err_n3w = (Te_n3w_err / Te_n3w) ** 2
    Te_mean_err = err_kw * 0
    if err_kw < 0.2:
        Te_mean_err += err_kw
    if err_n3w < 0.2:
        Te_mean_err += err_n3w
    Te_mean_err = np.sqrt(Te_mean_err) * Te_mean
    if te_key[st40.pulse] == "mean":
        Te_xrcs = Te_mean
        Te_xrcs_err = Te_mean_err
    elif te_key[st40.pulse] == "kw":
        Te_xrcs = Te_kw
        Te_xrcs_err = Te_kw_err
    elif te_key[st40.pulse] == "n3w":
        Te_xrcs = Te_n3w
        Te_xrcs_err = Te_n3w_err

    Ti_w_pos = _Ti_w_pos.mean("run")
    Ti_w_pos_err_in = _Ti_w_pos_err_in.max("run")
    Ti_w_pos_err_out = _Ti_w_pos_err_out.max("run")

    Ne_err = _Ne.std("run")
    Te_err = _Te.std("run")
    Ti_err = _Ti.std("run")

    plt.figure()
    plt.plot(Ti.rhop, Ti, color=COLORS["ion"], label="Ions")
    plt.fill_between(Ti.rhop, Ti - Ti_err, Ti + Ti_err, color=COLORS["ion"])
    plt.plot(Te.rhop, Te, color=COLORS["electron"], label="Electrons")
    plt.fill_between(Te.rhop, Te - Te_err, Te + Te_err, color=COLORS["electron"])

    plt.hlines(
        Te_xrcs,
        Ti_w_pos - Ti_w_pos_err_in,
        Ti_w_pos + Ti_w_pos_err_out,
        color="white",
        linewidth=plt.rcParams["lines.linewidth"] * 1.5,
    )
    plt.hlines(
        Te_xrcs,
        Ti_w_pos - Ti_w_pos_err_in,
        Ti_w_pos + Ti_w_pos_err_out,
        color=COLORS["electron"],
    )
    plt.vlines(
        Ti_w_pos,
        Te_xrcs - Te_xrcs_err,
        Te_xrcs + Te_xrcs_err,
        color="white",
        linewidth=plt.rcParams["lines.linewidth"] * 1.5,
    )
    plt.vlines(
        Ti_w_pos, Te_xrcs - Te_xrcs_err, Te_xrcs + Te_xrcs_err, color=COLORS["electron"]
    )
    plt.scatter(
        Ti_w_pos,
        Te_xrcs,
        marker="o",
        color=COLORS["electron"],
        facecolor="white",
        label="XRCS",
        zorder=3,
    )

    plt.hlines(
        Ti_w,
        Ti_w_pos - Ti_w_pos_err_in,
        Ti_w_pos + Ti_w_pos_err_out,
        color="white",
        linewidth=plt.rcParams["lines.linewidth"] * 1.5,
    )
    plt.hlines(
        Ti_w,
        Ti_w_pos - Ti_w_pos_err_in,
        Ti_w_pos + Ti_w_pos_err_out,
        color=COLORS["ion"],
    )
    plt.vlines(
        Ti_w_pos,
        Ti_w - Ti_w_err,
        Ti_w + Ti_w_err,
        color="white",
        linewidth=plt.rcParams["lines.linewidth"] * 1.5,
    )
    plt.vlines(Ti_w_pos, Ti_w - Ti_w_err, Ti_w + Ti_w_err, color=COLORS["ion"])
    plt.scatter(
        Ti_w_pos,
        Ti_w,
        marker="o",
        color=COLORS["ion"],
        facecolor="white",
        zorder=3,
    )

    Ti_cxff_pi_pos = Ti_cxff_pi_pos_all.mean("R_shift")
    Ti_cxff_pi_pos_err = Ti_cxff_pi_pos_all.max("R_shift") - Ti_cxff_pi_pos_all.min(
        "R_shift"
    )
    plt.hlines(
        Ti_cxff_pi,
        Ti_cxff_pi_pos - Ti_cxff_pi_pos_err,
        Ti_cxff_pi_pos + Ti_cxff_pi_pos_err,
        color="white",
        linewidth=plt.rcParams["lines.linewidth"] * 1.5,
    )
    plt.hlines(
        Ti_cxff_pi,
        Ti_cxff_pi_pos - Ti_cxff_pi_pos_err,
        Ti_cxff_pi_pos + Ti_cxff_pi_pos_err,
        color=COLORS["ion"],
    )
    plt.vlines(
        Ti_cxff_pi_pos,
        Ti_cxff_pi - Ti_cxff_pi_err,
        Ti_cxff_pi + Ti_cxff_pi_err,
        color="white",
        linewidth=plt.rcParams["lines.linewidth"] * 1.5,
    )
    plt.vlines(
        Ti_cxff_pi_pos,
        Ti_cxff_pi - Ti_cxff_pi_err,
        Ti_cxff_pi + Ti_cxff_pi_err,
        color=COLORS["ion"],
    )
    plt.scatter(
        Ti_cxff_pi_pos,
        Ti_cxff_pi,
        marker="s",
        color=COLORS["ion"],
        facecolor="white",
        label="CXRS",
        zorder=3,
    )
    set_axis_sci()
    plt.xlabel(r"$\rho_{pol}$")
    plt.ylabel("[keV]")
    plt.legend()
    plt.title(f"{st40.pulse} @ t={tplot:1.2f}s")
    if ylim is not None:
        plt.ylim(ylim)

    if save_fig:
        save_figure(FIG_PATH, f"{st40.pulse}_Stan_PPCF_figure", save_fig=save_fig)

    plt.figure()
    plt.vlines(tplot, 0, 1.0e4)
    st40.raw_data["xrcs"]["ti_w"].plot(marker="o", label="Ti_w")
    st40.raw_data["xrcs"]["te_kw"].plot(marker="o", label="Te_kw")
    st40.raw_data["xrcs"]["te_n3w"].plot(marker="o", label="Te_n3w")
    for chan in st40.raw_data["cxff_pi"]["ti"].sel(channel=cxff_pi_chans).channel:
        print(
            chan.values,
            Ti_cxff_pi.R.sel(channel=chan).values + np.mean(R_shift_scan),
            plasma[run].equilibrium.rmag.sel(t=tplot, method="nearest").values,
        )
        st40.raw_data["cxff_pi"]["ti"].sel(channel=chan).plot(
            marker="s", label="Ti_cxrs"
        )
    plt.legend()

    if write_csv:
        to_write = {
            "Rho-poloidal": Te.rhop.values,
            "Ne value (m**-3)": Ne.values,
            "Ne error (m**-3)": Ne_err.values,
            "Te value (eV)": Te.values,
            "Te error (eV)": Te_err.values,
            "Ti value (eV)": Ti.values,
            "Ti error (eV)": Ti_err.values,
        }
        df = pd.DataFrame(to_write)
        df.to_csv(f"{FIG_PATH}{st40.pulse}_{t.values:1.3f}s_HDA_profiles.csv")

    return

    # if plot or save_fig:
    #     load_modelling.plot_plasmas(
    #         plasma,
    #         tplot,
    #         code=code,
    #         save_fig=save_fig,
    #         fig_style=fig_style,
    #         alpha=alpha,
    #     )
    #
    #     load_modelling.plot_data_bckc_comparison(
    #         st40,
    #         bckc,
    #         plasma,
    #         tplot,
    #         code=code,
    #         save_fig=save_fig,
    #         fig_style=fig_style,
    #         alpha=alpha,
    #     )

    # return st40, bckc, plasma


def smmh1_evolution(data=None, savefig=False, ext="png"):

    # pulses = [9520, 9780, 9850, 10009]
    pulses = [10619, 10620, 10605, 10607]

    data = compare_pulses(
        pulses,
        data=data,
        qpop=[
            "nbi:pin",
            "mag:vloop",
            "smmh1:ne_bar",
            "halpha:brightness",
            "sxr_diode_1:brightness",
            "brems:brightness",
            "cxff_pi:ti",
            "cxff_pi:vtor",
            "xrcs:te_kw",
            "xrcs:ti_w",
            "mhd:ampl_odd_n",
        ],
        savefig=savefig,
        ext=ext,
        add_pulse_label_to="efit:ipla",
    )

    return data


def compare_pulses(
    pulses: list = [9538, 9780, 9783, 9831, 10014],
    tstart: float = 0.001,
    tend: float = 0.15,
    alpha: float = 0.9,
    xlabel: str = "Time (s)",
    figname="",
    savefig: bool = False,
    qpop: list = [""],
    data: list = None,
    R_cxrs: float = 0.4649,
    ext: str = "png",
    add_pulse_label_to="nbi:pin",
):
    """
    Compare time traces of different pulses
    for APS:
    compare_pulses(qpop=["lines:h_alpha", "mhd:ampl_odd_n"])
    """

    qdict = available_quantities()
    for q in qpop:
        if q in qdict.keys():
            qdict.pop(q)

    iax = -1
    set_plot_rcparams("time_evolution")

    hh = range(8000, 9677 + 1)
    dh = range(9685, 9787 + 1)
    dd = range(9802, 10050 + 1)
    pulse_labels = []
    for p in pulses:
        if p in hh:
            pulse_labels.append("HH")
        elif p in dh:
            pulse_labels.append("DH")
        elif p in dd:
            pulse_labels.append("DD")
        else:
            pulse_labels.append("")

    instruments = [q.split(":")[0] for q in qdict.keys()]
    if data is None:
        data = []
        for pulse in pulses:
            print(pulse)
            st40_data = ReadST40(pulse, tstart, tend)
            st40_data(instruments=instruments, raw_only=True)
            raw_data = st40_data.raw_data

            if "cxff_pi" in raw_data.keys():
                chan = np.argmin(np.abs(raw_data["cxff_pi"]["ti"].R - R_cxrs).values)
                _data = raw_data["cxff_pi"]["ti"].sel(channel=chan)
                _data = xr.where(_data > 0, _data, np.nan)
                _error = raw_data["cxff_pi"]["ti"].error.sel(channel=chan)
                _error = xr.where(_error > 0, _error, np.nan)
                raw_data["cxff_pi"]["ti"] = _data
                raw_data["cxff_pi"]["ti"] = raw_data["cxff_pi"]["ti"].assign_coords(
                    error=(raw_data["cxff_pi"]["ti"].dims, _error.data)
                )

                _data = raw_data["cxff_pi"]["vtor"].sel(channel=chan)
                _data = xr.where(_data > 0, _data, np.nan)
                _error = raw_data["cxff_pi"]["vtor"].error.sel(channel=chan)
                _error = xr.where(_error > 0, _error, np.nan)
                raw_data["cxff_pi"]["vtor"] = _data
                raw_data["cxff_pi"]["vtor"] = raw_data["cxff_pi"]["vtor"].assign_coords(
                    error=(raw_data["cxff_pi"]["vtor"].dims, _error.data)
                )

            if "hnbi1" in raw_data.keys():
                raw_data["nbi"] = {
                    "pin": raw_data["hnbi1"]["pin"] + raw_data["rfx"]["pin"]
                }
            if "efit" in raw_data.keys():
                smmh1_l = 2 * (raw_data["efit"]["rmjo"] - raw_data["efit"]["rmji"]).sel(
                    rhop=1
                )
                raw_data["smmh1"]["ne_bar"] = raw_data["smmh1"]["ne"] / smmh1_l.interp(
                    t=raw_data["smmh1"]["ne"].t
                )

            data.append(raw_data)

    fig, axs = plt.subplots(len(qdict.keys()), figsize=(6, 8))

    for key in qdict.keys():
        print(key)
        iax += 1
        ax = axs[iax]
        plot_quantity(
            st40_data,
            ax,
            data,
            key,
            pulses,
            qdict,
            add_pulse_label_to=add_pulse_label_to,
            alpha=alpha,
        )
        if iax != (len(axs) - 1):
            ax.xaxis.set_ticklabels([])
        else:
            ax.set_xlabel(xlabel)

    if savefig:
        name = ""
        for pulse in pulses:
            name += f"_{pulse}"
        if len(figname) > 1:
            name += f"_{figname}"
        save_figure(
            path_name=FIG_PATH, fig_name=f"time_evolution_comparison{name}", ext=ext
        )

    return data


def plot_quantity(
    st40_data: ReadST40,
    ax: plt.axis,
    data: list,
    qkey: str,
    pulses: list,
    qdict: dict,
    linestyle="solid",
    add_pulse_label_to: str = "",
    alpha: float = 0.9,
):

    ncols = 4
    if len(pulses) > ncols:
        ncols = len(pulses)
    cols = CMAP(np.linspace(0.1, 0.75, ncols, dtype=float))

    if "label" not in qdict[qkey]:
        qdict[qkey]["label"] = ""
    if "marker" not in qdict[qkey].keys():
        qdict[qkey]["marker"] = ""

    diag, quant = qkey.split(":")
    label = ""
    add_pulse_label = False
    if qkey == add_pulse_label_to:
        add_pulse_label = True
    for i in range(len(data)):
        skip = False
        if diag not in data[i].keys():
            skip = True
            continue
        if quant not in data[i][diag].keys():
            skip = True
            continue

        if add_pulse_label:
            label = str(pulses[i])
        val = data[i][diag][quant] * qdict[qkey]["const"]

        if "error" in data[i][diag][quant].attrs.keys():
            err = data[i][diag][quant].error * qdict[qkey]["const"]
            ind = np.where(np.isfinite(val.values) * np.isfinite(err.values))[0]
            ax.errorbar(
                val.t[ind],
                val.values[ind],
                err.values[ind],
                alpha=alpha,
                color=cols[i],
                marker=qdict[qkey]["marker"],
                label=label,
                linestyle=linestyle,
            )
        else:
            ind = np.where(np.isfinite(val.values))[0]
            ax.plot(
                val.t[ind],
                val.values[ind],
                alpha=alpha,
                color=cols[i],
                label=label,
                marker=qdict[qkey]["marker"],
                linestyle=linestyle,
            )

    if skip:
        return

    ax.set_ylabel(qdict[qkey]["ylabel"])

    # add pulses label
    if add_pulse_label:
        loc = "upper right"
        plt.gca()
        ax.legend(frameon=True, handlelength=None, loc=loc)

    # add quantity label
    if len(qdict[qkey]["label"]) > 0:
        ax_label = ax.twinx()
        ax_label.plot(
            [np.nan],
            [np.nan],
            label=qdict[qkey]["label"],
        )
        ax_label.get_yaxis().set_visible(False)
        ax_label.legend(frameon=False, handlelength=0, loc="upper left")

    ax.set_xlim(st40_data.tstart, st40_data.tend)
    if "ylim" in qdict[qkey].keys():
        ylim = qdict[qkey]["ylim"]
    ax.set_ylim(ylim)


def available_quantities():

    qdict = {
        "efit:ipla": {
            "const": 1.0e-6,
            "label": "$I_P$ $EFIT$",
            "ylabel": "$(MA)$",
            "ylim": (0, None),
        },
        "efit:wp": {
            "const": 1.0e-3,
            "label": "$W_P$ $EFIT$",
            "ylabel": "$(kJ)$",
            "ylim": (0, None),
        },
        "nbi:pin": {
            "const": 1.0e-6,
            "label": "$P_{NBI}$",
            "ylabel": "$(MW)$",
            "ylim": (0, None),
        },
        "mag:vloop": {
            "const": 1.0,
            "ylabel": "$V_{loop}$ $(V)$",
            "ylim": (0, None),
        },
        "smmh1:ne_bar": {
            "const": 1.0e-19,
            "label": r"$\overline{N}_e$ $SMM$",
            "ylabel": "($10^{19}$ $m^{-3}$)",
            "ylim": (0, 7),
        },
        "smmh1:ne": {
            "const": 1.0e-19,
            "label": "$N_e$-int SMM",
            "ylabel": "($10^{19}$ $m^{-2}$)",
            "ylim": (0, 7),
        },
        "nirh1:ne": {
            "const": 1.0e-19,
            "label": "$N_e$-int NIR",
            "ylabel": "($10^{19}$ $m^{-2}$)",
            "ylim": (0, 15),
        },
        "halpha:brightness": {
            "const": 1.0,
            "label": r"$H_\alpha$ $Filter$",
            "ylabel": r"($a.u.)$",
            "ylim": (0, 0.5),
        },
        "brems:brightness": {
            "const": 1.0,
            "label": r"$Bremsstrahlung $Filter$",
            "ylabel": r"($a.u.)$",
            "ylim": (0, 0.5),
        },
        "sxr_diode_1:brightness": {
            "const": 1.0e-3,
            "label": "$P_{SXR}$",
            "ylabel": "$(a.u.)$",
            "ylim": (0, None),
        },
        "xrcs:te_kw": {
            "const": 1.0e-3,
            "label": "$T_e$ $XRCS$",
            "ylabel": "$(keV)$",
            "ylim": (0, None),
            "error": True,
            "marker": "o",
        },
        "xrcs:ti_w": {
            "const": 1.0e-3,
            "label": "$T_i$ $XRCS$",
            "ylabel": "$(keV)$",
            "ylim": (0, 10),
            "error": True,
            "marker": "o",
        },
        "cxff_pi:ti": {
            "const": 1.0e-3,
            "label": "$T_i$ $CXRS$",
            "ylabel": "$(keV)$",
            "ylim": (0, 10),
            "error": True,
            "marker": "x",
        },
        "cxff_pi:vtor": {
            "const": 1.0e-3,
            "label": "$V_{tor}$ $CXRS$",
            "ylabel": "$(km/s)$",
            "ylim": (0, 300),
            "error": True,
            "marker": "x",
        },
        "mhd:ampl_odd_n": {
            "const": 1.0,
            "label": "$Odd$ $n$ $MHD$",
            "ylabel": "$(a.u.)$",
            "ylim": (0, None),
        },
    }

    return qdict



import numpy as np

from indica.converters import LineOfSightTransform
from indica.converters import TransectCoordinates


def cxrs_transform_example(nchannels: int):
    x_positions = np.linspace(0.2, 0.8, nchannels)
    y_positions = np.linspace(0.0, 0.0, nchannels)
    z_positions = np.linspace(0.0, 0.0, nchannels)

    transect_transform = TransectCoordinates(
        x_positions,
        y_positions,
        z_positions,
        "pi",
        machine_dimensions=((0.15, 0.95), (-0.7, 0.7)),
    )
    return transect_transform


def helike_transform_example(nchannels):
    los_end = np.full((nchannels, 3), 0.0)
    los_end[:, 0] = 0.17
    los_end[:, 1] = 0.0
    los_end[:, 2] = np.linspace(0.2, -0.5, nchannels)
    los_start = np.array([[0.9, 0, 0]] * los_end.shape[0])
    los_start[:, 2] = -0.1
    origin = los_start
    direction = los_end - los_start

    los_transform = LineOfSightTransform(
        origin[0:nchannels, 0],
        origin[0:nchannels, 1],
        origin[0:nchannels, 2],
        direction[0:nchannels, 0],
        direction[0:nchannels, 1],
        direction[0:nchannels, 2],
        name="xrcs",
        machine_dimensions=((0.15, 0.95), (-0.7, 0.7)),
        passes=1,
    )
    return los_transform


def interferometer_transform_example(nchannels):
    los_start = np.array([[0.8, 0, 0]]) * np.ones((nchannels, 3))
    los_start[:, 2] = np.linspace(0, -0.2, nchannels)
    los_end = np.array([[0.17, 0, 0]]) * np.ones((nchannels, 3))
    los_end[:, 2] = np.linspace(0, -0.2, nchannels)
    origin = los_start
    direction = los_end - los_start
    los_transform = LineOfSightTransform(
        origin[:, 0],
        origin[:, 1],
        origin[:, 2],
        direction[:, 0],
        direction[:, 1],
        direction[:, 2],
        name="smmh1",
        machine_dimensions=((0.15, 0.95), (-0.7, 0.7)),
        passes=2,
    )
    return los_transform


def ts_transform_example(nchannels):
    x_positions = np.linspace(0.2, 0.8, nchannels)
    y_positions = np.linspace(0.0, 0.0, nchannels)
    z_positions = np.linspace(0.0, 0.0, nchannels)
    transform = TransectCoordinates(
        x_positions,
        y_positions,
        z_positions,
        "ts",
        machine_dimensions=((0.15, 0.95), (-0.7, 0.7)),
    )
    return transform
